from policyengine_us.model_api import *


class wi_retirement_income_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Wisconsin retirement income subtraction from federal AGI"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.wi.gov/TaxForms2021/2021-ScheduleSB.pdf"
        "https://www.revenue.wi.gov/TaxForms2021/2021-ScheduleSB-inst.pdf#page=9"
        "https://www.revenue.wi.gov/TaxForms2022/2022-ScheduleSBf.pdf"
        "https://www.revenue.wi.gov/TaxForms2022/2022-ScheduleSB-Inst.pdf#page=7"
    )
    defined_for = "wi_retirement_income_subtraction_agi_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.wi.tax.income
        psri = p.subtractions.retirement_income
        person = tax_unit.members
        age = person("age", period)
        age_eligible = age >= psri.min_age
        retirement_income = person("taxable_pension_income", period)
        head_or_spouse = ~person("is_tax_unit_dependent", period)
        uncapped_retinc = retirement_income * age_eligible * head_or_spouse
        capped_retinc = min_(psri.max_amount, uncapped_retinc)
        return tax_unit.sum(capped_retinc)



from policyengine_us.model_api import *


class medical_out_of_pocket_expenses(Variable):
    value_type = float
    entity = Person
    label = "Medical out of pocket expenses"
    unit = USD
    definition_period = YEAR
    adds = [
        "health_insurance_premiums",
        "other_medical_expenses",
        # Note: Excludes over_the_counter_health_expenses
        # as IRS does not include them in the itemized deduction, and
        # USDA only includes doctor-approved over-the-counter medications
        # in their medical out-of-pocket expenses definition for SNAP.
    ]



from policyengine_us.model_api import *


class ma_child_and_family_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Massachusetts child and family tax credit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.mass.gov/info-details/massachusetts-child-and-family-tax-credit"
    defined_for = StateCode.MA

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.ma.tax.income.credits.child_and_family
        person = tax_unit.members
        dependent = person("is_tax_unit_dependent", period)
        age = person("age", period)
        child = age < p.child_age_limit
        elderly = age >= p.elderly_age_limit
        disabled = person("is_disabled", period)
        eligible = dependent & (child | elderly | disabled)
        count_eligible = tax_unit.sum(eligible)
        capped_eligible = min_(count_eligible, p.dependent_cap)
        return capped_eligible * p.amount



from policyengine_us.model_api import *


class savers_credit(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Retirement Savings Credit"
    unit = USD
    reference = (
        "https://www.irs.gov/pub/irs-pdf/f8880.pdf",
        "https://www.law.cornell.edu/uscode/text/26/25B#c",
    )

    def formula(tax_unit, period, parameters):
        credit_limit = tax_unit("savers_credit_credit_limit", period)
        potential = tax_unit("savers_credit_potential", period)
        return min_(credit_limit, potential)



from policyengine_us.model_api import *


class co_collegeinvest_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Colorado collegeinvest subtraction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.colorado.gov/sites/tax/files/documents/DR0104AD_2022.pdf#page=1",
        "https://tax.colorado.gov/sites/tax/files/documents/DR_104_Book_2022.pdf#page=12",
        "https://law.justia.com/codes/colorado/2022/title-39/article-22/part-1/section-39-22-104/",
        # C.R.S. 39-22-104(4)(i)(II)(B)
    )
    defined_for = StateCode.CO

    def formula(tax_unit, period, parameters):
        investment_amount = tax_unit("investment_in_529_plan", period)
        p = parameters(
            period
        ).gov.states.co.tax.income.subtractions.collegeinvest_contribution
        cap = p.max_amount[tax_unit("filing_status", period)]
        return min_(investment_amount, cap)



from policyengine_us.model_api import *


class co_sales_tax_refund(Variable):
    value_type = float
    entity = TaxUnit
    label = "Colorado sales tax refund"
    unit = USD
    definition_period = YEAR
    reference = "https://tax.colorado.gov/sites/tax/files/documents/DR_0104_Book_2022.pdf#page=23"
    defined_for = "co_sales_tax_refund_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.co.tax.income.credits.sales_tax_refund.amount
        filing_status = tax_unit("filing_status", period)
        multiplier = p.multiplier[filing_status]
        if p.flat_amount_enabled:
            amount = p.amount
        else:
            agi = tax_unit("co_modified_agi", period)
            multiplier = p.multiplier[filing_status]
            amount = p.scale.calc(agi)
        return multiplier * amount



from policyengine_us.model_api import *


class dc_tanf_immigration_status_eligible_person(Variable):
    value_type = bool
    entity = Person
    label = "Eligible person for DC Temporary Assistance for Needy Families (TANF) based on immigration status"
    definition_period = MONTH
    reference = (
        "https://code.dccouncil.gov/us/dc/council/code/sections/4-205.24#(a)"
    )
    defined_for = StateCode.DC

    def formula(person, period, parameters):
        p = parameters(period).gov.states.dc.dhs
        immigration_status = person("immigration_status", period)
        immigration_status_str = immigration_status.decode_to_str()
        is_citizen = (
            immigration_status == immigration_status.possible_values.CITIZEN
        )
        qualified_noncitizen = np.isin(
            immigration_status_str,
            p.qualified_noncitizen_statuses,
        )
        return qualified_noncitizen | is_citizen



from policyengine_us.model_api import *


class ks_count_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    label = "number of KS exemptions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.ksrevenue.gov/pdf/ip21.pdf"
        "https://www.ksrevenue.gov/pdf/ip22.pdf"
    )
    defined_for = StateCode.KS

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        statuses = filing_status.possible_values
        joint = filing_status == statuses.JOINT
        hoh = filing_status == statuses.HEAD_OF_HOUSEHOLD
        adults = where(joint | hoh, 2, 1)
        dependents = tax_unit("tax_unit_dependents", period)
        return adults + dependents



from policyengine_us.model_api import *


class de_standard_deduction_joint(Variable):
    value_type = float
    entity = Person
    label = "Delaware standard deduction when married filing jointly"
    unit = USD
    definition_period = YEAR
    reference = "https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf#page=8"
    defined_for = StateCode.DE

    def formula(person, period, parameters):
        base_sd = person("de_base_standard_deduction_joint", period)
        add_sd = person("de_additional_standard_deduction", period)
        head = person("is_tax_unit_head", period)
        add_sd_attributed_to_head = person.tax_unit.sum(add_sd) * head
        return base_sd + add_sd_attributed_to_head



from policyengine_us.model_api import *


class ma_tafdc_non_financial_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for Massachusetts Temporary Assistance for Families with Dependent Children (TAFDC)"
    definition_period = MONTH
    reference = (
        "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-703-010"
    )
    defined_for = StateCode.MA

    def formula(spm_unit, period, parameters):
        immigration_eligible = spm_unit(
            "ma_tafdc_immigration_status_eligible", period
        )
        dependent_criteria_eligible = spm_unit(
            "ma_tafdc_dependent_criteria_eligible", period
        )
        return immigration_eligible & dependent_criteria_eligible



from policyengine_us.model_api import *


class mt_tuition_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Montana tuition subtraction"
    definition_period = YEAR
    reference = "https://leg.mt.gov/bills/mca/title_0150/chapter_0300/part_0210/section_0200/0150-0300-0210-0200.html"
    defined_for = StateCode.MT

    adds = ["mt_tuition_subtraction_person"]



from policyengine_us.model_api import *


class va_non_refundable_eitc_if_claimed(Variable):
    value_type = float
    entity = TaxUnit
    label = "Virginia non-refundable EITC if claimed"
    unit = USD
    definition_period = YEAR
    reference = "https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf#page=32"
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        federal_eitc = tax_unit("eitc", period)
        p = parameters(period).gov.states.va.tax.income.credits.eitc.match
        # The filer can either claim the non-refundable EITC or the low income credit.
        va_non_refundable_eitc = p.non_refundable * federal_eitc
        low_income_credit = tax_unit("va_low_income_tax_credit", period)
        return max_(va_non_refundable_eitc, low_income_credit)



from policyengine_us.model_api import *


class form_4972_lumpsum_distributions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Lump-sum distributions reported on IRS Form 4972"
    unit = USD
    definition_period = YEAR
    documentation = "Lump-sum distributions reported on IRS Form 4972."



from policyengine_us.model_api import *


class ia_exemption_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Iowa exemption credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.iowa.gov/sites/default/files/2021-12/IA6251%2841131%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-01/IA6251%2841131%29.pdf"
    )
    defined_for = StateCode.IA

    def formula(tax_unit, period, parameters):
        # count adult and dependent exemptions
        adult_count = tax_unit("head_spouse_count", period)
        filing_status = tax_unit("filing_status", period)
        hoh_status = filing_status.possible_values.HEAD_OF_HOUSEHOLD
        hoh_bonus = filing_status == hoh_status
        dependent_count = tax_unit("tax_unit_dependents", period)
        # count extra adult exemptions based on being elderly and/or blind
        p = parameters(period).gov.states.ia.tax.income
        exemption = p.credits.exemption
        elder_head = tax_unit("age_head", period) >= exemption.elderly_age
        elder_spouse = tax_unit("age_spouse", period) >= exemption.elderly_age
        elder_count = elder_head.astype(int) + elder_spouse.astype(int)
        blind_head = tax_unit("blind_head", period)
        blind_spouse = tax_unit("blind_spouse", period)
        blind_count = blind_head.astype(int) + blind_spouse.astype(int)
        additional_count = elder_count + blind_count
        return (
            (adult_count + hoh_bonus) * exemption.personal
            + additional_count * exemption.additional
            + dependent_count * exemption.dependent
        )



from policyengine_us.model_api import *


class mi_retirement_benefits_deduction_tier_three_ss_exempt_not_retired_eligible_people(
    Variable
):
    value_type = int
    entity = TaxUnit
    label = "Number of eligible people for the Michigan non-retired tier three retirement benefits deduction"
    definition_period = YEAR
    reference = (
        "http://legislature.mi.gov/doc.aspx?mcl-206-30",  # (9)(d)
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf#page=17",
        "https://www.michigan.gov/taxes/iit/retirement-and-pension-benefits",
    )
    defined_for = StateCode.MI

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        #  Recipients should receive retirement benefits from employment exempt from Social Security
        has_retirement_benefits_from_ss_exempt_employment = (
            person(
                "retirement_benefits_from_ss_exempt_employment",
                period,
            )
            > 0
        )

        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)

        eligible_people = (
            has_retirement_benefits_from_ss_exempt_employment
            * is_head_or_spouse
        )

        return tax_unit.sum(eligible_people)



from policyengine_us.model_api import *


class tax_unit_taxable_social_security(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Taxable Social Security benefits"
    documentation = "Social security (OASDI) benefits included in AGI, including tier 1 railroad retirement benefits."
    unit = USD
    reference = "https://www.law.cornell.edu/uscode/text/26/86"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.irs.social_security.taxability
        gross_ss = tax_unit("tax_unit_social_security", period)

        # The legislation directs the usage an income definition that is
        # a particularly modified AGI, plus half of gross Social Security
        # payments. We assume that the 'half' here is the same underlying
        # parameter as the lower taxability marginal rate (also 50% in the
        # baseline), and that they would be mechanically the same parameter.

        combined_income = tax_unit(
            "tax_unit_combined_income_for_social_security_taxability", period
        )
        filing_status = tax_unit("filing_status", period)
        status = filing_status.possible_values
        separate = filing_status == status.SEPARATE
        cohabitating = tax_unit("cohabitating_spouses", period)
        # Cohabitating married couples filing separately receive a base
        # and adjusted base amount of 0
        base_amount = where(
            separate & cohabitating,
            p.threshold.base.separate_cohabitating,
            p.threshold.base.main[filing_status],
        )
        adjusted_base_amount = where(
            separate & cohabitating,
            p.threshold.adjusted_base.separate_cohabitating,
            p.threshold.adjusted_base.main[filing_status],
        )

        under_first_threshold = combined_income < base_amount
        under_second_threshold = combined_income < adjusted_base_amount

        combined_income_excess = tax_unit(
            "tax_unit_ss_combined_income_excess", period
        )
        excess_over_adjusted_base = max_(
            0, combined_income - adjusted_base_amount
        )

        amount_if_under_second_threshold = p.rate.base * min_(
            combined_income_excess, gross_ss
        )
        amount_if_over_second_threshold = min_(
            p.rate.additional * excess_over_adjusted_base
            + min_(
                amount_if_under_second_threshold,
                p.rate.base * (adjusted_base_amount - base_amount),
            ),
            p.rate.additional * gross_ss,
        )
        return select(
            [
                under_first_threshold,
                under_second_threshold,
                True,
            ],
            [
                0,
                amount_if_under_second_threshold,
                amount_if_over_second_threshold,
            ],
        )



from policyengine_us.model_api import *


class meets_tanf_non_cash_gross_income_test(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Meets gross income test for TANF non-cash benefit"
    documentation = "Income eligibility (gross income as a percent of the poverty line) for TANF non-cash benefit for SNAP BBCE"
    definition_period = MONTH

    def formula(spm_unit, period, parameters):
        state = spm_unit.household("state_code_str", period)
        # All limits and incomes here expressed as % of FPG.
        limits = parameters(period).gov.hhs.tanf.non_cash.income_limit
        gross_limit = limits.gross[state]
        gross_income = spm_unit("snap_gross_income_fpg_ratio", period)
        return gross_income <= gross_limit



from policyengine_us.model_api import *


class md_itemized_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maryland itemized deductions"
    unit = USD
    definition_period = YEAR
    reference = [
        "https://govt.westlaw.com/mdc/Document/N05479690A64A11DBB5DDAC3692B918BC?viewType=FullText&originationContext=documenttoc&transitionType=CategoryPageItem&contextData=(sc.Default)",
        "https://www.marylandtaxes.gov/forms/21_forms/Resident_Booklet.pdf#page=5",
        "https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf#page=5",
        "https://mgaleg.maryland.gov/2025RS/Chapters_noln/CH_604_hb0352e.pdf#page=167",  # Maryland House Bill 352 - Budget Reconciliation and Financing Act of 2025
    ]
    defined_for = StateCode.MD

    def formula(tax_unit, period, parameters):
        # Base itemized deductions (federal itemized deductions less SALT plus capped property taxes)
        base_itemized = add(
            tax_unit,
            period,
            ["itemized_deductions_less_salt", "capped_property_taxes"],
        )

        # Apply income-based phase-out if applicable
        p = parameters(period).gov.states.md.tax.income.deductions.itemized
        if p.phase_out.applies:
            filing_status = tax_unit("filing_status", period)
            md_agi = tax_unit("adjusted_gross_income", period)

            # Calculate phase-out reduction
            excess_income = max_(
                md_agi - p.phase_out.threshold[filing_status], 0
            )
            phase_out_reduction = excess_income * p.phase_out.rate

            # Apply phase-out (itemized deductions cannot go below zero)
            return max_(base_itemized - phase_out_reduction, 0)

        # When phase-out does not apply, no phase-out
        return base_itemized



from policyengine_core.populations import Population
from policyengine_core.simulations import Simulation


def get_stored_variables(simulation: Simulation) -> list:
    stored_variables = []
    for variable in simulation.tax_benefit_system.variables:
        holder = simulation.get_holder(variable)
        if len(holder.get_known_periods()) > 0:
            stored_variables.append(variable)
    return stored_variables


class BranchedSimulation:
    def __init__(self, population: Population):
        self.simulation = population.simulation

    def __enter__(self):
        simulation = self.simulation.clone()
        simulation.max_spiral_loops = 10
        simulation._check_for_cycle = lambda *args: None
        self.computed_variables = get_stored_variables(simulation)
        simulation.tracer = self.simulation.tracer
        self.branched_simulation: Simulation = simulation

        return self.branched_simulation

    def __exit__(self, type, value, traceback):
        added_variables = set(
            get_stored_variables(self.branched_simulation)
        ) - set(self.computed_variables)
        for variable in added_variables:
            self.simulation.get_holder(variable).delete_arrays()



from policyengine_us.model_api import *


class railroad_benefits(Variable):
    value_type = float
    entity = Person
    label = "Recieves any railroad benefits"
    definition_period = YEAR
    default_value = 0



from policyengine_us.model_api import *
from policyengine_core.periods import instant


class ny_ctc_pre_2024_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "NY CTC pre-2024 eligibility"
    documentation = (
        "Whether the tax unit is eligible for NY CTC under pre-2024 rules"
    )
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/606"  # (c-1)
    defined_for = StateCode.NY

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ny.tax.income.credits.ctc

        # Only eligible if post-2024 rules are NOT in effect
        if p.post_2024.in_effect:
            return False

        person = tax_unit.members
        age = person("age", period)

        # Check if we have qualifying children
        if not p.pre_tcja:
            qualifies_for_federal_ctc = person("ctc_qualifying_child", period)
        else:
            # Initialize pre-TCJA CTC branch for eligibility check
            simulation = tax_unit.simulation
            pre_tcja_ctc = simulation.get_branch("pre_tcja_ctc")
            pre_tcja_ctc.tax_benefit_system = (
                simulation.tax_benefit_system.clone()
            )
            branch_parameters = pre_tcja_ctc.tax_benefit_system.parameters
            # Update parameters to pre-TCJA values.
            for (
                ctc_parameter
            ) in branch_parameters.gov.irs.credits.ctc.get_descendants():
                if isinstance(ctc_parameter, Parameter):
                    ctc_parameter.update(
                        start=instant("2017-01-01"),
                        stop=instant("2035-01-01"),
                        value=ctc_parameter("2017-01-01"),
                    )
            # Delete all arrays from pre-TCJA CTC branch.
            for variable in pre_tcja_ctc.tax_benefit_system.variables:
                if "ctc" in variable:
                    pre_tcja_ctc.delete_arrays(variable)
            qualifies_for_federal_ctc = pre_tcja_ctc.person(
                "ctc_qualifying_child", period
            )

        qualifies = qualifies_for_federal_ctc & (age >= p.minimum_age)
        qualifying_children = tax_unit.sum(qualifies)
        return qualifying_children > 0



from policyengine_us.model_api import *


class is_ma_income_tax_exempt(Variable):
    value_type = bool
    entity = TaxUnit
    label = "MA income tax exempt"
    unit = USD
    definition_period = YEAR
    reference = "https://malegislature.gov/Laws/GeneralLaws/PartI/TitleIX/Chapter62/Section5"
    defined_for = StateCode.MA

    def formula(tax_unit, period, parameters):
        agi = tax_unit("ma_agi", period)
        threshold = tax_unit("ma_income_tax_exemption_threshold", period)
        return agi <= threshold



from policyengine_us.model_api import *


class mn_cdcc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Minnesota child and dependent care expense credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.state.mn.us/sites/default/files/2023-02/m1cd_21.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1cd_22_0.pdf"
    )
    defined_for = StateCode.MN

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.mn.tax.income.credits.cdcc
        person = tax_unit.members
        # determine eligibility for Minnesota CDCC
        if p.separate_filers_excluded:
            filing_status = tax_unit("filing_status", period)
            eligible = filing_status != filing_status.possible_values.SEPARATE
        else:
            eligible = True
        # calculate number of qualifying dependents
        # ... children
        age = person("age", period)
        qualifies_by_age = age < p.child_age
        # ... disability
        non_head = ~person("is_tax_unit_head", period)
        disabled = person("is_incapable_of_self_care", period)
        qualifies_by_disability = non_head & disabled
        dep_count = tax_unit.sum(qualifies_by_age | qualifies_by_disability)
        # calculate qualifying care expenses
        expense = tax_unit("tax_unit_childcare_expenses", period)
        # ... cap expense by number of qualifying dependents
        eligible_count = min_(dep_count, p.maximum_dependents)
        expense = min_(expense, p.maximum_expense * eligible_count)
        # ... cap expense by lower earnings of head and spouse if present
        expense = min_(expense, tax_unit("min_head_spouse_earned", period))
        # calculate pre-phaseout credit amount
        agi = tax_unit("adjusted_gross_income", period)
        pre_phaseout_amount = expense * p.expense_fraction.calc(agi)
        # calculate post-phaseout credit amount
        excess_agi = max_(0, agi - p.phaseout_threshold)
        phaseout_amount = excess_agi * p.phaseout_rate
        amount = max_(0, pre_phaseout_amount - phaseout_amount)
        # credit amount only for eligibles
        return eligible * amount



from policyengine_us.model_api import *


class de_taxable_income_indv(Variable):
    value_type = float
    entity = Person
    label = (
        "Delaware taxable income when married couples are filing separately"
    )
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.DE

    def formula(person, period, parameters):
        agi = person("de_agi_indiv", period)
        deductions = person("de_deduction_indv", period)
        return max_(0, agi - deductions)



from policyengine_us.model_api import *


def create_adjust_income_limit_and_min_children_by_filing_status() -> Reform:
    class nyc_school_tax_credit_eligible(Variable):
        value_type = bool
        entity = TaxUnit
        label = "Eligible for NYC School Tax Credit"
        definition_period = YEAR
        defined_for = "in_nyc"

        def formula(tax_unit, period, parameters):
            # Get the NYC School Tax Credit reform part of the parameter tree.
            p = parameters(period).gov.contrib.local.nyc.stc

            # Get income that counts towards the NYC School Tax Credit.
            nyc_stc_income = tax_unit("nyc_school_credit_income", period)

            # Get the tax unit's filing status.
            filing_status = tax_unit("filing_status", period)

            # Calculate eligibility.
            income_eligible = nyc_stc_income <= p.income_limit[filing_status]
            children_eligible = (
                tax_unit("tax_unit_children", period) >= p.min_children
            )

            return income_eligible & children_eligible

    class nyc_school_tax_credit(Variable):
        value_type = float
        entity = TaxUnit
        label = "NYC School Tax Credit"
        unit = USD
        definition_period = YEAR
        defined_for = "nyc_school_tax_credit_eligible"

        adds = [
            "nyc_school_tax_credit_fixed_amount",
            "nyc_school_tax_credit_rate_reduction_amount",
        ]

    class reform(Reform):
        def apply(self):
            self.update_variable(nyc_school_tax_credit_eligible)
            self.update_variable(nyc_school_tax_credit)

    return reform


def create_adjust_income_limit_by_filing_status_and_eligibility_by_children_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_adjust_income_limit_and_min_children_by_filing_status()

    p = parameters(period).gov.contrib.local.nyc.stc

    if p.adjust_income_limit_by_filing_status_and_eligibility_by_children:
        return create_adjust_income_limit_and_min_children_by_filing_status()
    else:
        return None


adjust_income_limit_and_min_children_by_filing_status = create_adjust_income_limit_by_filing_status_and_eligibility_by_children_reform(
    None, None, bypass=True
)



from policyengine_us.model_api import *


class ne_child_care_subsidy_eligible_child(Variable):
    value_type = bool
    entity = Person
    label = "Nebraska Child Care Subsidy program eligible child"
    definition_period = YEAR
    reference = (
        "https://nebraskalegislature.gov/laws/statutes.php?statute=68-1206",
        "https://dhhs.ne.gov/Pages/Child-Care-Subsidy-Information-for-Parents.aspx",
    )
    defined_for = StateCode.NE

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.states.ne.dhhs.child_care_subsidy.age_threshold
        age = person("age", period)
        has_special_needs = person("ne_dhhs_has_special_needs", period)
        age_threshold = where(has_special_needs, p.special_needs, p.base)
        return age <= age_threshold



from policyengine_us.model_api import *


class ca_ala_general_assistance_countable_income(Variable):
    value_type = float
    entity = SPMUnit
    unit = USD
    definition_period = MONTH
    label = "Eligible for Alameda County General Assistance based on age requirements"
    defined_for = "in_ala"
    reference = "https://www.alamedacountysocialservices.org/acssa-assets/PDF/GA-Policies/GA-Regulations.pdf#page=29"

    adds = ["ca_ala_general_assistance_countable_income_person"]



from policyengine_us.model_api import *


class il_total_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "Illinois total tax"
    unit = USD
    definition_period = YEAR

    defined_for = StateCode.IL

    adds = ["il_income_tax_before_refundable_credits", "il_use_tax"]



from policyengine_us.model_api import *


class mn_niit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Minnesota Net Investment Income Tax (NIIT)"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.state.mn.us/sites/default/files/2024-12/m1m-24.pdf"
        "https://www.revisor.mn.gov/statutes/cite/290.033"
    )
    defined_for = StateCode.MN

    def formula(tax_unit, period, parameters):
        """
        Computes Minnesota's net investment income tax using a bracket structure.
        Currently, the brackets are defined such that:
          - Amounts up to $1,000,000 are taxed at 0%
          - Amounts above $1,000,000 are taxed at 1%
        """
        p = parameters(period).gov.states.mn.tax.income.niit
        # Minnesota defines its NIIT base following the IRC, except exempting
        # gains from some agricultural/rural/forested land. We do not account
        # for this exemption.
        net_investment_income = tax_unit("net_investment_income", period)
        return p.rate.calc(net_investment_income)



from policyengine_us.model_api import *


class ami(Variable):
    value_type = float
    entity = Household
    label = "Area median income"
    documentation = "Area median income for a four-person household"
    definition_period = YEAR

    def formula(household, period, parameters):
        # Only calculate for LA County and Denver County for now. Otherwise zero.
        in_la = household("in_la", period)
        in_denver = household("in_denver", period)
        # https://www.hcd.ca.gov/sites/default/files/docs/grants-and-funding/income-limits-2023.pdf
        LA_COUNTY_AMI = 98_200
        DENVER_COUNTY_AMI = 124_100
        return in_la * LA_COUNTY_AMI | in_denver * DENVER_COUNTY_AMI



from policyengine_us.model_api import *


class mi_retirement_benefits_deduction_tier_three(Variable):
    value_type = float
    entity = TaxUnit
    label = "Michigan retirement benefits deduction for tier three"
    unit = USD
    definition_period = YEAR
    reference = (
        "http://legislature.mi.gov/doc.aspx?mcl-206-30",  # (9)(c) & (d)
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf#page=17",
        "https://www.michigan.gov/taxes/iit/retirement-and-pension-benefits",
    )
    defined_for = "mi_retirement_benefits_deduction_tier_three_eligible"

    def formula(tax_unit, period, parameters):
        ss_exempt_not_retired_amount = tax_unit(
            "mi_retirement_benefits_deduction_tier_three_ss_exempt_not_retired",
            period,
        )
        ss_exempt_retired_amount = tax_unit(
            "mi_retirement_benefits_deduction_tier_three_ss_exempt_retired",
            period,
        )

        return max_(ss_exempt_not_retired_amount, ss_exempt_retired_amount)



from policyengine_us.model_api import *


class co_non_refundable_ctc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Non-refundable Child Tax Credit replicated to include the Colorado limitations"
    unit = USD
    documentation = (
        "Total value of the non-refundable portion of the Child Tax Credit."
    )
    definition_period = YEAR
    reference = (
        # C.R.S. 39-22-129. Child tax credit - legislative declaration - definitions.
        "https://casetext.com/statute/colorado-revised-statutes/title-39-taxation/specific-taxes/income-tax/article-22-income-tax/part-1-general/section-39-22-129-child-tax-credit-legislative-declaration-definitions-repeal",
        # 2022 Colorado Child Tax Credit
        "https://tax.colorado.gov/sites/tax/files/documents/DR_0104CN_2022.pdf#page=1",
        # Colorado Individual Income Tax Filing Guide - Instructions for Select Credits from the DR 0104CR - Line 1 Child Tax Credit
        "https://tax.colorado.gov/sites/tax/files/documents/DR_104_Book_2022.pdf#page=16",
    )
    defined_for = StateCode.CO

    def formula(tax_unit, period, parameters):
        # follow 2022 DR 0104CN form and its instructions (in Book cited above):
        maximum = tax_unit("co_federal_ctc_maximum", period)  # Line 3
        limiting_tax_liability = tax_unit(
            "ctc_limiting_tax_liability", period
        )  # Line 4 - 6
        return min_(maximum, limiting_tax_liability)  # Line 7



from policyengine_us.model_api import *


class TenureType(Enum):
    RENTED = "rented"
    OWNED_OUTRIGHT = "owned outright"
    OWNED_WITH_MORTGAGE = "owned with mortgage"
    NONE = "none"


class tenure_type(Variable):
    value_type = Enum
    entity = Household
    possible_values = TenureType
    default_value = TenureType.NONE
    label = "tenure type"
    definition_period = YEAR



from policyengine_us.model_api import *


class il_tanf(Variable):
    value_type = float
    entity = SPMUnit
    label = "Illinois Temporary Assistance for Needy Families (TANF)"
    unit = USD
    definition_period = MONTH
    reference = "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-112.250"
    defined_for = "il_tanf_eligible"

    def formula(spm_unit, period, parameters):
        payment_level = spm_unit(
            "il_tanf_payment_level_for_grant_calculation", period
        )
        countable_income = spm_unit(
            "il_tanf_countable_income_for_grant_calculation", period
        )
        return max_(payment_level - countable_income, 0)



from policyengine_us.model_api import *


class qualified_furnace_or_hot_water_boiler_expenditures(Variable):
    value_type = float
    entity = TaxUnit
    label = "Expenditures on qualified natural gas, propane, or oil furnaces or hot water boilers"
    documentation = "Must achieve an annual fuel utilization efficiency rate of not less than 95."
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/25C#d_4"



from policyengine_us.model_api import *


class mn_nonrefundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Minnesota nonrefundable income tax credits"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1c_21.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1c_22.pdf"
    )
    defined_for = StateCode.MN
    adds = "gov.states.mn.tax.income.credits.nonrefundable"



from policyengine_us.model_api import *


class ca_state_supplement_food_allowance_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "California SSI state supplement food allowance eligible"
    definition_period = MONTH
    defined_for = StateCode.CA
    reference = "https://leginfo.legislature.ca.gov/faces/codes_displaySection.xhtml?lawCode=WIC&sectionNum=12200"

    def formula(spm_unit, period, parameters):
        living_arrangements_allow_for_food_preparation = spm_unit.household(
            "living_arrangements_allow_for_food_preparation", period
        )
        aged_disabled_count = spm_unit(
            "ca_state_supplement_aged_disabled_count", period
        )
        return (
            aged_disabled_count > 0
        ) & ~living_arrangements_allow_for_food_preparation



from policyengine_us.model_api import *


class ca_cvrp(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    label = "California Clean Vehicle Rebate Project"
    unit = USD
    documentation = (
        "Total California Clean Vehicle Rebate Project (CVRP) benefit"
    )
    reference = "https://cleanvehiclerebate.org/en/eligibility-guidelines"
    defined_for = StateCode.CA

    def formula(person, period, parameters):
        # Calculate normal rebate.
        normal_eligible = person("is_ca_cvrp_normal_rebate_eligible", period)
        vehicle_amount = person("ca_cvrp_vehicle_rebate_amount", period)
        normal_amount = normal_eligible * vehicle_amount
        # Calculate increased rebate (more means-tested).
        p_increased_amount = parameters(
            period
        ).gov.states.ca.calepa.carb.cvrp.increased_rebate.amount
        increased_eligible = person(
            "is_ca_cvrp_increased_rebate_eligible", period
        )
        bought_qualifying_ev = vehicle_amount > 0
        increased_amount = (
            increased_eligible & bought_qualifying_ev
        ) * p_increased_amount
        return normal_amount + increased_amount



from policyengine_us.model_api import *


class ny_itemized_deductions_reduction_based_on_charitable_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = (
        "New York itemized deductions reduction based on charitable deduction"
    )
    unit = USD
    definition_period = YEAR
    reference = "https://www.tax.ny.gov/pdf/2024/inc/it196i_2024.pdf#page=20"  # Line 46, worksheet 5 & 6
    defined_for = "ny_itemized_deductions_reduction_based_on_charitable_deduction_applies"

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.ny.tax.income.deductions.itemized.reduction
        agi = tax_unit("ny_agi", period)
        charitable_deduction = tax_unit("charitable_deduction", period)
        applicable_rate = p.charitable_deduction_rate.calc(agi)
        max_deduction = tax_unit("ny_itemized_deductions_max", period)
        total_reduction_amount = charitable_deduction * applicable_rate
        return max_(max_deduction - total_reduction_amount, 0)



from policyengine_us.model_api import *


class equiv_household_net_income(Variable):
    value_type = float
    entity = Household
    label = "equivalised net income"
    definition_period = YEAR
    unit = USD

    def formula(household, period, parameters):
        number_of_people = household("household_size", period)
        net_income = household("household_net_income", period)
        return net_income / (number_of_people**0.5)



from policyengine_us.model_api import *


class il_ccap_parent_meets_working_requirements(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Parent meets Illinois Child Care Assistance Program (CCAP) working requirements"
    definition_period = MONTH
    defined_for = StateCode.IL
    reference = "https://www.dhs.state.il.us/page.aspx?item=104995"

    def formula(spm_unit, period, parameters):
        person = spm_unit.members
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        is_working = (
            add(
                person,
                period,
                ["employment_income", "self_employment_income"],
            )
            > 0
        )
        is_full_time_student = person("is_full_time_student", period)
        ineligible_parent = head_or_spouse & ~(
            is_working | is_full_time_student
        )
        return spm_unit.sum(ineligible_parent) == 0



from policyengine_us.model_api import *


class la_federal_tax_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Louisiana federal tax deduction"
    unit = USD
    definition_period = YEAR
    reference = [
        "https://revenue.louisiana.gov/TaxForms/IT540iWEB(2022)D1.pdf#page=2",  # 2022 repealed
        "https://revenue.louisiana.gov/TaxForms/IT540i(2021)%20Instructions.pdf#page=3",  # 2021 line 9
        "https://law.justia.com/codes/louisiana/2021/revised-statutes/title-47/rs-298/",  # (3)
    ]
    defined_for = StateCode.LA

    def formula(tax_unit, period, parameters):
        tax_before_refundable_credits = tax_unit(
            "income_tax_before_refundable_credits", period
        )
        # The deduction is also reduced by the ACA PTC repayment amount.
        # We do not model this as it depends on the advance PTC received throughout the year.
        reductions = add(
            tax_unit,
            period,
            ["form_4972_lumpsum_distributions"],
        )
        amount = max_(0, tax_before_refundable_credits - reductions)

        # The deduction was discontinued in 2022
        p = parameters(period).gov.states.la.tax.income.deductions.federal_tax
        return amount * p.availability



from policyengine_us.model_api import *


class ny_itemized_deductions_max(Variable):
    value_type = float
    entity = TaxUnit
    label = "NY uncapped itemized deductions"
    unit = USD
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/615"
    defined_for = StateCode.NY
    adds = [
        "itemized_deductions_less_salt",
        "real_estate_taxes",
        "ny_college_tuition_deduction",
    ]



from policyengine_us.model_api import *


class tax_unit_is_filer(Variable):
    value_type = float
    entity = TaxUnit
    label = "files taxes"
    unit = USD
    documentation = (
        "Whether this tax unit has a non-zero income tax liability."
    )
    definition_period = YEAR

    """
    (a) General rule
    Returns with respect to income taxes under subtitle A shall be made by the following:
    (1)
    (A) Every individual having for the taxable year gross income which equals or exceeds the exemption amount, except that a return shall not be required of an individual—
    (i) who is not married (determined by applying section 7703), is not a surviving spouse (as defined in section 2(a)), is not a head of a household (as defined in section 2(b)), and for the taxable year has gross income of less than the sum of the exemption amount plus the basic standard deduction applicable to such an individual,
    (ii) who is a head of a household (as so defined) and for the taxable year has gross income of less than the sum of the exemption amount plus the basic standard deduction applicable to such an individual,
    (iii) who is a surviving spouse (as so defined) and for the taxable year has gross income of less than the sum of the exemption amount plus the basic standard deduction applicable to such an individual, or
    (iv) who is entitled to make a joint return and whose gross income, when combined with the gross income of his spouse, is, for the taxable year, less than the sum of twice the exemption amount plus the basic standard deduction applicable to a joint return, but only if such individual and his spouse, at the close of the taxable year, had the same household as their home.
    Clause (iv) shall not apply if for the taxable year such spouse makes a separate return or any other taxpayer is entitled to an exemption for such spouse under section 151(c).
    (B) The amount specified in clause (i), (ii), or (iii) of subparagraph (A) shall be increased by the amount of 1 additional standard deduction (within the meaning of section 63(c)(3)) in the case of an individual entitled to such deduction by reason of section 63(f)(1)(A) (relating to individuals age 65 or more), and the amount specified in clause (iv) of subparagraph (A) shall be increased by the amount of the additional standard deduction for each additional standard deduction to which the individual or his spouse is entitled by reason of section 63(f)(1).
    (C) The exception under subparagraph (A) shall not apply to any individual—
    (i) who is described in section 63(c)(5) and who has—
    (I) income (other than earned income) in excess of the sum of the amount in effect under section 63(c)(5)(A) plus the additional standard deduction (if any) to which the individual is entitled, or
    (II) total gross income in excess of the standard deduction, or
    (ii) for whom the standard deduction is zero under section 63(c)(6).
    (D) For purposes of this subsection—
    (i) The terms "standard deduction", "basic standard deduction" and "additional standard deduction" have the respective meanings given such terms by section 63(c).
    (ii) The term "exemption amount" has the meaning given such term by section 151(d). In the case of an individual described in section 151(d)(2), the exemption amount shall be zero.
    """

    def formula(tax_unit, period, parameters):
        gross_income = add(tax_unit, period, ["irs_gross_income"])
        exemption_amount = parameters(period).gov.irs.income.exemption.amount

        # (a)(1)(A), (a)(1)(B)

        filing_status = tax_unit("filing_status", period).decode_to_str()
        separate = filing_status == "SEPARATE"
        standard_deduction = tax_unit("standard_deduction", period)
        threshold = where(
            separate,
            exemption_amount,
            standard_deduction + exemption_amount,
        )

        income_over_exemption_amount = gross_income > threshold

        # (a)(1)(C)

        unearned_income_threshold = 500 + tax_unit(
            "additional_standard_deduction", period
        )
        unearned_income = gross_income - add(
            tax_unit, period, ["earned_income"]
        )
        unearned_income_over_threshold = (
            unearned_income > unearned_income_threshold
        )

        required_to_file = (
            income_over_exemption_amount | unearned_income_over_threshold
        )

        tax_refund = tax_unit("income_tax", period) < 0
        not_required_but_likely_filer = ~required_to_file & tax_refund

        # (a)(1)(D) is just definitions

        return required_to_file | not_required_but_likely_filer



from policyengine_us.model_api import *


class ia_regular_tax_consolidated(Variable):
    value_type = float
    entity = TaxUnit
    label = "Iowa regular tax for years on or after 2023"
    unit = USD
    definition_period = YEAR
    reference = "https://revenue.iowa.gov/media/2748/download?inline"
    defined_for = StateCode.IA

    def formula(tax_unit, period, parameters):
        taxable_income = tax_unit("ia_taxable_income_consolidated", period)
        p = parameters(period).gov.states.ia.tax.income.rates
        filing_status = tax_unit("filing_status", period)
        joint = filing_status == filing_status.possible_values.JOINT
        return where(
            joint,
            p.by_filing_status.joint.calc(taxable_income),
            p.by_filing_status.other.calc(taxable_income),
        )



from policyengine_us.model_api import *


class co_denver_property_tax_relief_income(Variable):
    value_type = float
    entity = SPMUnit
    unit = USD
    label = "Denver Property Tax Relief income"
    definition_period = YEAR
    reference = "https://denvergov.org/files/assets/public/v/2/denver-human-services/documents/property-tax-relief/dptr-instructions-2023.pdf#page=1"

    adds = "gov.local.co.denver.dhs.property_tax_relief.income_sources"



from policyengine_us.model_api import *


class nj_eitc_income_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "New Jersey Eligible for EITC"
    definition_period = YEAR
    reference = "https://law.justia.com/codes/new-jersey/2022/title-54a/section-54a-4-7/"
    defined_for = StateCode.NJ

    def formula(tax_unit, period, parameters):
        # Get parameter tree for federal EITC.
        p = parameters(period).gov.irs.credits.eitc

        # Check if they are above the investment income limit.
        no_loss_capital_gains = max_(0, tax_unit("net_capital_gains", period))
        eitc_investment_income = (
            add(
                tax_unit,
                period,
                ["net_investment_income", "tax_exempt_interest_income"],
            )
            # replace limited-loss capital gains with no-loss capital gains
            - tax_unit("loss_limited_net_capital_gains", period)
            + no_loss_capital_gains
        )
        inv_income_disqualified = (
            eitc_investment_income > p.phase_out.max_investment_income
        )

        # Determine if filer is above federal EITC income threshold.
        # We can find the income threshold by looking at their filing status and the phase-out rate for 0 children.
        # Note: this assumes the same phaseout rate for all filing statuses (which is the case).
        is_joint = tax_unit("tax_unit_is_joint", period)
        joint_bonus = p.phase_out.joint_bonus.calc(0)
        phaseout_rate = p.phase_out.rate.calc(0)
        phaseout_start = p.phase_out.start.calc(0) + is_joint * joint_bonus
        max_credit = p.max.calc(0)
        completed_phaseout = max_credit / phaseout_rate + phaseout_start
        return (
            tax_unit("adjusted_gross_income", period) < completed_phaseout
        ) & ~inv_income_disqualified



from policyengine_us.model_api import *


class al_retirement_exemption_person(Variable):
    value_type = float
    entity = Person
    label = "Alabama retirement exemption for each person"
    unit = USD
    # Alabama Schedule RS Part II & III Line 10, Alabama Form 40 Booklet Page 14 Pension & Annuities, Alabama Section 40-18-19 (a)(13)
    documentation = (
        "https://www.revenue.alabama.gov/wp-content/uploads/2024/01/23schrsinstr.pdf#page=1"
        "https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2024/01/23f40bk.pdf#page=14"
        "https://casetext.com/statute/code-of-alabama/title-40-revenue-and-taxation/chapter-18-income-taxes/article-1-general-provisions/section-40-18-19-exemptions-generally"
    )
    definition_period = YEAR
    defined_for = "al_retirement_exemption_eligible_person"

    def formula(person, period, parameters):
        p = parameters(period).gov.states.al.tax.income.exemptions.retirement
        retirement_income = add(
            person,
            period,
            ["taxable_retirement_distributions", "taxable_pension_income"],
        )
        return min_(retirement_income, p.cap)



from policyengine_us.model_api import *


class de_additional_standard_deduction(Variable):
    value_type = float
    entity = Person
    label = "Delaware additional standard deduction"
    unit = USD
    definition_period = YEAR
    reference = "https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf#page=8"
    defined_for = StateCode.DE

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.states.de.tax.income.deductions.standard.additional
        age_eligible = (person("age", period) >= p.age_threshold).astype(int)
        blind_eligible = person("is_blind", period).astype(int)

        return (age_eligible + blind_eligible) * p.amount



from policyengine_us.model_api import *


class ny_tanf_resources_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "New York TANF resources eligible"
    definition_period = YEAR
    defined_for = StateCode.NY

    def formula(spm_unit, period, parameters):
        # The amount of assets that a family may own and qualify for FA is $2,000
        # except for households in which any member is age 60 or over in which case $3,000 in assets can be owned.
        # https://otda.ny.gov/policy/tanf/TANF-State-Plan-2021-2023.pdf#page=7, #10
        person = spm_unit.members
        p = parameters(period).gov.states.ny.otda.tanf.eligibility.resources
        person_meets_higher_resource_limit_age = (
            person("age", period) >= p.higher_resource_limit_age_threshold
        )
        has_anyone_with_higher_resource_limit_age = spm_unit.any(
            person_meets_higher_resource_limit_age
        )
        resource_limit = where(
            has_anyone_with_higher_resource_limit_age,
            p.higher_limit,
            p.lower_limit,
        )
        countable_resources = spm_unit("ny_tanf_countable_resources", period)
        return countable_resources <= resource_limit



from policyengine_us.model_api import *


class has_disabled_spouse(Variable):
    value_type = bool
    entity = Person
    label = "person's marriage partner in JOINT filing unit is disabled"
    definition_period = YEAR

    def formula(person, period, parameters):
        filing_status = person.tax_unit("filing_status", period)
        married = filing_status == filing_status.possible_values.JOINT
        head_disabled = person.tax_unit("head_is_disabled", period)
        spouse_disabled = person.tax_unit("spouse_is_disabled", period)
        is_head = person("is_tax_unit_head", period)
        is_spouse = person("is_tax_unit_spouse", period)
        return married & (
            (is_head & spouse_disabled) | (is_spouse & head_disabled)
        )



from policyengine_us.model_api import *


class nj_other_retirement_special_exclusion(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Jersey Other Retirement Special Exclusion"
    unit = USD
    documentation = "New Jersey other retirement special exclusion"
    definition_period = YEAR
    reference = (
        "https://www.state.nj.us/treasury/taxation/pdf/current/1040i.pdf#page=21",
        "https://law.justia.com/codes/new-jersey/2022/title-54a/section-54a-6-15/",
    )
    defined_for = StateCode.NJ

    def formula(tax_unit, period, parameters):
        # If filer (and spouse, if joint) are never eligible for social security benefits, they are eligible for special exclusion.
        # Get the pension/retirement exclusion portion of the parameter tree.
        p = parameters(period).gov.states.nj.tax.income.exclusions.retirement

        person = tax_unit.members
        is_head = person("is_tax_unit_head", period)
        is_spouse = person("is_tax_unit_spouse", period)
        filing_status = tax_unit("filing_status", period)
        status = filing_status.possible_values
        joint = filing_status == status.JOINT
        never_eligible_ss = person(
            "never_eligible_for_social_security_benefits", period
        )

        head_eligible = tax_unit.sum(is_head * never_eligible_ss)
        spouse_eligible = tax_unit.sum(is_spouse * never_eligible_ss)

        eligible = where(
            joint,
            head_eligible * spouse_eligible,
            head_eligible,
        )

        return p.special_exclusion.amount[filing_status] * eligible



from policyengine_us.model_api import *


class ctc_child_individual_maximum_arpa(Variable):
    value_type = float
    entity = Person
    label = "CTC maximum amount (child under ARPA)"
    unit = USD
    documentation = "The CTC entitlement in respect of this person as a child, under the American Rescue Plan Act."
    definition_period = YEAR
    defined_for = "ctc_qualifying_child"

    def formula(person, period, parameters):
        age = person("age", period)
        return parameters(period).gov.irs.credits.ctc.amount.arpa.calc(age)



from policyengine_us.model_api import *


class nj_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Jersey refundable credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NJ
    adds = [
        "nj_property_tax_credit",
        "nj_eitc",
        "nj_cdcc",
        "nj_ctc",
    ]



from policyengine_us.model_api import *


class md_refundable_cdcc(Variable):
    value_type = float
    entity = TaxUnit
    label = "MD refundable CDCC"
    documentation = "Maryland refundable Child and Dependent Care Tax Credit"
    unit = USD
    definition_period = YEAR
    reference = "https://casetext.com/statute/code-of-maryland/article-tax-general/title-10-income-tax/subtitle-7-income-tax-credits/section-10-716-for-child-care-or-dependent-care"
    defined_for = StateCode.MD

    def formula(tax_unit, period, parameters):
        # Note: The MD refundable CDCC is based on the MD CDCC and MD tax before credits
        # *without respect to the other non-refundable credits* like MD EITC.
        filing_status = tax_unit("filing_status", period)
        agi = tax_unit("adjusted_gross_income", period)
        p = parameters(period).gov.states.md.tax.income.credits.cdcc
        cdcc = tax_unit("md_cdcc", period)
        # Refundable has its own AGI cap.
        cap = p.eligibility.refundable_agi_cap[filing_status]
        eligible = agi <= cap
        tax_before_credits = tax_unit("md_income_tax_before_credits", period)
        return eligible * max_(0, cdcc - tax_before_credits)



from policyengine_us.model_api import *


class weekly_hours_worked(Variable):
    value_type = float
    entity = Person
    label = "average weekly hours worked"
    unit = "hour"
    documentation = "Hours worked per week on average."
    definition_period = YEAR
    adds = [
        "weekly_hours_worked_before_lsr",
        "weekly_hours_worked_behavioural_response",
    ]


class weekly_hours_worked_before_lsr(Variable):
    value_type = float
    entity = Person
    label = "average weekly hours worked (before labor supply responses)"
    unit = "hour"
    definition_period = YEAR
    default_value = 40


class weekly_hours_worked_behavioural_response_income_elasticity(Variable):
    value_type = float
    entity = Person
    label = "behavioural response in weekly hours worked (income effect)"
    unit = "hour"
    definition_period = YEAR

    def formula(person, period, parameters):
        original = person("weekly_hours_worked_before_lsr", period)
        lsr = person("labor_supply_behavioral_response", period)

        if (lsr != 0).any():
            income_effect = person("income_elasticity_lsr", period)
        else:
            income_effect = np.zeros_like(original)

        original_emp = person("employment_income_before_lsr", period)
        original_self_emp = person("self_employment_income_before_lsr", period)
        original_earnings = original_emp + original_self_emp

        lsr_relative_change = np.divide(
            income_effect,
            original_earnings,
            # Assign no LSR change to people with no original earnings to avoid dividing by zero.
            out=np.zeros_like(income_effect, dtype=np.float32),
            where=original_earnings != 0,
        )

        return original * lsr_relative_change


class weekly_hours_worked_behavioural_response_substitution_elasticity(
    Variable
):
    value_type = float
    entity = Person
    label = "behavioural response in weekly hours worked (substitution effect)"
    unit = "hour"
    definition_period = YEAR

    def formula(person, period, parameters):
        original = person("weekly_hours_worked_before_lsr", period)
        lsr = person("labor_supply_behavioral_response", period)
        if (lsr != 0).any():
            substitution_effect = person("substitution_elasticity_lsr", period)
        else:
            substitution_effect = np.zeros_like(original)
        original_emp = person("employment_income_before_lsr", period)
        original_self_emp = person("self_employment_income_before_lsr", period)
        original_earnings = original_emp + original_self_emp

        lsr_relative_change = np.divide(
            substitution_effect,
            original_earnings,
            # Assign no LSR change to people with no original earnings to avoid dividing by zero.
            out=np.zeros_like(substitution_effect, dtype=np.float32),
            where=original_earnings != 0,
        )

        return original * lsr_relative_change


class weekly_hours_worked_behavioural_response(Variable):
    value_type = float
    entity = Person
    label = "behavioural response in weekly hours worked"
    unit = "hour"
    definition_period = YEAR
    adds = [
        "weekly_hours_worked_behavioural_response_income_elasticity",
        "weekly_hours_worked_behavioural_response_substitution_elasticity",
    ]



from policyengine_us.model_api import *


class mi_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Michigan Earned Income Tax Credit"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MI

    def formula(tax_unit, period, parameters):
        federal_eitc = tax_unit("eitc", period)
        p = parameters(period).gov.states.mi.tax.income.credits.eitc
        return federal_eitc * p.match



from policyengine_us.model_api import *


class income_tax_non_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "federal non-refundable income tax credits"
    unit = USD
    adds = "gov.irs.credits.non_refundable"



from policyengine_us.model_api import *


class nc_demographic_tanf_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    definition_period = YEAR
    label = "North Carolina Demographic eligibility for TANF"
    documentation = "Whether any person in a family applying for the Temporary Assistance for Needy Families program meets demographic requirements."
    defined_for = StateCode.NC

    def formula(spm_unit, period, parameters):
        age_limit = parameters(
            period
        ).gov.states.nc.ncdhhs.tanf.need_standard.age_limit
        person = spm_unit.members
        is_child = person("age", period) < age_limit

        household_size = spm_unit("nc_tanf_household_size", period)

        return spm_unit.any(is_child) & (household_size > 0)



from policyengine_us.model_api import *


class pa_cdcc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Pennsylvania Child and Dependent Care Credit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40dc.pdf"  # 2022 form
    defined_for = StateCode.PA

    def formula(tax_unit, period, parameters):
        # Get the federal CDCC value
        # Pennsylvania matches the potential federal credit
        cdcc = tax_unit("cdcc_potential", period)
        # Access the parameter path
        p = parameters(period).gov.states.pa.tax.income.credits.cdcc
        # Multiply the federal CDCC value by the Pennsylvania CDCC rate
        return p.match * cdcc



from policyengine_us.model_api import *


class share_of_care_and_support_costs_paid_by_tax_filer(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    label = "The percentage of care and support costs of a senior paid by the tax filer"

    def formula(person, period, parameters):
        care_and_support_payment = person(
            "care_and_support_payments_from_tax_filer", period
        )
        care_and_support_costs = person("care_and_support_costs", period)
        support_payment_ratio = np.zeros_like(care_and_support_costs)
        mask = care_and_support_costs != 0
        support_payment_ratio[mask] = (
            care_and_support_payment[mask] / care_and_support_costs[mask]
        )
        return support_payment_ratio



from policyengine_us.model_api import *


class dc_gac_income_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = (
        "Eligible for DC General Assistance for Children (GAC) due to income"
    )
    definition_period = MONTH
    reference = (
        "https://code.dccouncil.gov/us/dc/council/code/sections/4-205.10"
    )
    defined_for = StateCode.DC

    def formula(spm_unit, period, parameters):
        countable_income = spm_unit("dc_gac_countable_income", period)
        standard_payment = spm_unit("dc_gac_standard_payment", period)
        return countable_income <= standard_payment



from policyengine_us.model_api import *


class nj_childless_eitc_age_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "New Jersey Eligible for EITC"
    definition_period = YEAR
    reference = "https://law.justia.com/codes/new-jersey/2022/title-54a/section-54a-4-7/"
    defined_for = StateCode.NJ

    def formula(tax_unit, period, parameters):
        # Return True if all federal EITC conditions are met, except with modified age paramaters and household has no children.
        # Check if filing status is separate.
        filing_status = tax_unit("filing_status", period)
        separate = filing_status == filing_status.possible_values.SEPARATE
        # Check if tax unit has any EITC qualifying children.
        person = tax_unit.members
        no_qualifying_children = tax_unit("eitc_child_count", period) == 0

        # Get the NJ EITC paramaeter tree.
        p = parameters(period).gov.states.nj.tax.income.credits.eitc

        # Check if the filer meets NJ EITC age requirements.
        age = person("age", period)
        age_eligible = age >= p.eligibility.age.min

        return (
            ~separate
            & no_qualifying_children
            & tax_unit.any(age_eligible)
            & tax_unit("nj_eitc_income_eligible", period)
        )



from policyengine_us.model_api import *


class nj_cdcc(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Jersey CDCC"
    documentation = "New Jersey Child and Dependent Care Credit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.state.nj.us/treasury/taxation/pdf/current/1040i.pdf#page=44"
    defined_for = StateCode.NJ

    def formula(tax_unit, period, parameters):
        # Get New Jersey CDCC rate
        p = parameters(period).gov.states.nj.tax.income.credits.cdcc.rate

        # Get NJ taxable income
        taxable_income = tax_unit("nj_taxable_income", period)

        # Get federal CDCC
        # New Jersey matches the potential federal credit
        federal_cdcc = tax_unit("cdcc_potential", period)

        # Calculate NJ CDCC
        rate = p.calc(taxable_income, right=True)

        # Calculate total NJ CDCC
        return federal_cdcc * rate



from policyengine_us.model_api import *


class me_property_tax_fairness_credit_benefit_base(Variable):
    value_type = float
    entity = TaxUnit
    unit = USD
    label = "Maine property tax fairness credit benefit base"
    definition_period = YEAR
    defined_for = StateCode.ME

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.me.tax.income.credits.fairness.property_tax
        dependents = tax_unit("ctc_qualifying_children", period)
        filing_status = tax_unit("filing_status", period)
        filing_statuses = filing_status.possible_values
        single = filing_status == filing_statuses.SINGLE
        joint = filing_status == filing_statuses.JOINT
        joint_no_child = joint & (dependents == 0)
        head_of_household = filing_status == filing_statuses.HEAD_OF_HOUSEHOLD
        head_of_household_one_child = head_of_household & (
            dependents <= p.dependent_count_threshold
        )
        joint_with_child = joint & (dependents >= p.dependent_count_threshold)
        head_of_household_multiple_child = head_of_household & (
            dependents > p.dependent_count_threshold
        )
        # legal code does not mention surviving spouse, but it does appear in the tax form
        surviving_spouse = filing_status == filing_statuses.SURVIVING_SPOUSE
        # Separate filers are ineligible so not included in this select statement.
        # This benefit base does not include the senior amount that takes effect in 2024.
        return select(
            [
                single,
                joint_no_child | head_of_household_one_child,
                joint_with_child
                | head_of_household_multiple_child
                | surviving_spouse,
            ],
            [
                p.benefit_base.single,
                p.benefit_base.head_of_household_one_child,
                p.benefit_base.joint_or_head_of_household_multiple_children,
            ],
            default=0,
        )



from policyengine_us.model_api import *


class vita_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Eligible for the VITA program"
    definition_period = YEAR

    def formula(person, period, parameters):
        p = parameters(period).gov.irs.vita.eligibility
        income = person("irs_gross_income", period)
        is_eligible_income = income <= p.income_limit
        is_disabled = person("is_disabled", period)
        proficient_in_english = person("is_english_proficient", period)
        return is_eligible_income | is_disabled | ~proficient_in_english



from policyengine_us.model_api import *


class mt_dependent_exemptions_person(Variable):
    value_type = int
    entity = Person
    label = "Montana dependent exemption for each dependent"
    unit = USD
    definition_period = YEAR
    reference = "https://regulations.justia.com/states/montana/department-42/chapter-42-15/subchapter-42-15-4/rule-42-15-403/"
    defined_for = StateCode.MT

    def formula(person, period, parameters):
        p = parameters(period).gov.states.mt.tax.income.exemptions

        if p.applies:
            # Qualifying child under IRC 152(c), which defines for the EITC
            qualifying_child = person("is_child_dependent", period)
            # Disabled dependents get an additional exemption.
            disabled = person("is_disabled", period)

            eligible_dependent = qualifying_child * (1 + disabled)
            return eligible_dependent * p.amount

        return 0



from policyengine_us.model_api import *


class va_aged_blind_exemption_person(Variable):
    value_type = float
    entity = Person
    label = "Virginia aged/blind exemption for each person"
    defined_for = StateCode.VA
    unit = USD
    definition_period = YEAR
    reference = (
        "https://law.lis.virginia.gov/vacodefull/title58.1/chapter3/article2/"
    )

    def formula(person, period, parameters):
        p = parameters(period).gov.states.va.tax.income.exemptions
        aged_person = person("is_irs_aged", period).astype(int)
        blind_person = person("is_blind", period).astype(int)
        aged_blind_count = aged_person + blind_person
        return aged_blind_count * p.aged_blind



from policyengine_us.model_api import *


class de_files_separately(Variable):
    value_type = bool
    entity = TaxUnit
    label = "married couple files separately on the Delaware tax return"
    unit = USD
    definition_period = YEAR
    reference = "https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf"
    defined_for = StateCode.DE

    def formula(tax_unit, period, parameters):
        itax_indiv = add(
            tax_unit,
            period,
            ["de_income_tax_before_non_refundable_credits_indv"],
        )
        itax_joint = add(
            tax_unit,
            period,
            ["de_income_tax_before_non_refundable_credits_joint"],
        )
        return itax_indiv < itax_joint



from policyengine_us.model_api import *


class oh_uniformed_services_retirement_income_deduction(Variable):
    value_type = float
    entity = Person
    label = "Ohio Uniformed services retirement income"
    definition_period = YEAR
    unit = USD
    reference = "https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-bundle.pdf#page=4"
    defined_for = StateCode.OH



from policyengine_us.model_api import *


class il_income_tax_before_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Illinois income tax before refundable credits"
    unit = USD
    definition_period = YEAR

    defined_for = StateCode.IL

    def formula(tax_unit, period, parameters):
        taxbc = tax_unit("il_income_tax_before_non_refundable_credits", period)
        non_refundable_credits = tax_unit("il_nonrefundable_credits", period)
        return max_(taxbc - non_refundable_credits, 0)



from policyengine_us.model_api import *


class nc_ctc(Variable):
    value_type = float
    entity = TaxUnit
    label = "North Carolina credit for children"
    definition_period = YEAR
    unit = USD
    reference = "https://www.ncdor.gov/taxes-forms/individual-income-tax/credit-children"
    defined_for = StateCode.NC

    def formula(tax_unit, period, parameters):
        ctc_qualifying_children = tax_unit("ctc_qualifying_children", period)
        income = tax_unit("adjusted_gross_income", period)
        filing_status = tax_unit("filing_status", period)
        p = parameters(period).gov.states.nc.tax.income.credits.ctc
        status = filing_status.possible_values
        credit_amount = select(
            [
                filing_status == status.SINGLE,
                filing_status == status.HEAD_OF_HOUSEHOLD,
                filing_status == status.JOINT,
                filing_status == status.SURVIVING_SPOUSE,
                filing_status == status.SEPARATE,
            ],
            [
                p.single.calc(income),
                p.head_of_household.calc(income),
                p.joint.calc(income),
                p.surviving_spouse.calc(income),
                p.separate.calc(income),
            ],
        )
        return ctc_qualifying_children * credit_amount



from policyengine_us.model_api import *


class az_agi(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arizona adjusted gross income"
    unit = USD
    definition_period = YEAR
    documentation = "https://azdor.gov/sites/default/files/2023-03/FORMS_INDIVIDUAL_2020_140NRBOOKLET.pdf#page=18"  # Line 52
    defined_for = StateCode.AZ

    def formula(tax_unit, period, parameters):
        adjusted_gross_income = tax_unit("adjusted_gross_income", period)
        additions = tax_unit("az_additions", period)
        subtractions = tax_unit("az_subtractions", period)
        exemptions = tax_unit("az_exemptions", period)

        return max_(
            0, adjusted_gross_income + additions - subtractions - exemptions
        )



from policyengine_us.model_api import *


class la_general_relief_base_amount(Variable):
    value_type = float
    entity = SPMUnit
    unit = USD
    label = "Los Angeles County General Relief base amount"
    definition_period = MONTH
    defined_for = "la_general_relief_eligible"
    reference = "https://drive.google.com/file/d/1Oc7UuRFxJj-eDwTeox92PtmRVGnG9RjW/view?usp=sharing"

    def formula(spm_unit, period, parameters):
        # Is married is defined for the family,
        # the `> 0` is necessary for the np.where statement
        married = add(spm_unit, period, ["is_married"]) > 0
        people_eligible_based_on_immigration_status = add(
            spm_unit,
            period,
            ["la_general_relief_immigration_status_eligible_person"],
        )
        p = parameters(period).gov.local.ca.la.general_relief
        base_amount = where(
            married & (people_eligible_based_on_immigration_status == 2),
            p.amount.married,
            p.amount.single,
        )
        # The base amount phases out for recipients
        net_income = spm_unit("la_general_relief_net_income", period)
        excess_net_income = max_(net_income - p.phase_out.start, 0)
        phase_out_width = p.phase_out.max - p.phase_out.start
        reduction_percent = min_(excess_net_income / phase_out_width, 1)
        recipient = spm_unit("la_general_relief_recipient", period)
        # LA only reduces the amount for recipients, not new registrants.
        reduction = recipient * base_amount * reduction_percent
        return max_(base_amount - reduction, 0)



from policyengine_us.model_api import *


class nc_scca_age_group(Variable):
    value_type = int
    entity = Person
    label = "North Carolina SCCA age group"
    definition_period = YEAR
    reference = "https://ncchildcare.ncdhhs.gov/Portals/0/documents/pdf/M/Market_Rates_Centers_Eff_10-1.pdf?ver=9w52alSPhmrmo0N9gGVMEw%3d%3d"

    def formula(person, period, parameters):
        """
        Returns the age group for NC SCCA:
        1 - Infant (0 to <1)
        2 - Toddler (1 to <3)
        3 - Preschooler (3 to <6)

        The school age group is hardcoded as 4 (one more than the maximum value in the parameter)
        """
        age = person("age", period)
        is_school_age = person("nc_scca_is_school_age", period)
        p = parameters(period).gov.states.nc.ncdhhs.scca

        # School age group is 4 (one more than the maximum value in the age.group parameter)
        SCHOOL_AGE_GROUP = 4

        # Use standard calculation for non-school age, SCHOOL_AGE_GROUP for school age
        return where(is_school_age, SCHOOL_AGE_GROUP, p.age.group.calc(age))



from policyengine_us.model_api import *


class mt_elderly_homeowner_or_renter_credit_net_household_income(Variable):
    value_type = float
    entity = Person
    label = (
        "Net household income for Montana elderly homeowner or renter credit"
    )
    unit = USD
    definition_period = YEAR
    defined_for = "mt_elderly_homeowner_or_renter_credit_eligible"

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.states.mt.tax.income.credits.elderly_homeowner_or_renter
        # Only one claim is allowed per household
        # married taxpayer who are living apart may qualify for only one credit per year
        standard_exclusion = p.net_household_income.standard_exclusion
        # Allocate the income to the head
        head = person("is_tax_unit_head", period)
        gross_household_income = add(
            person.tax_unit,
            period,
            ["mt_elderly_homeowner_or_renter_credit_gross_household_income"],
        )
        gross_household_income_head = gross_household_income * head
        reduced_household_income = max_(
            gross_household_income_head - standard_exclusion, 0
        )
        reduction_rate = p.net_household_income.reduction_rate.calc(
            reduced_household_income
        )
        return reduced_household_income * reduction_rate



from policyengine_us.model_api import *


class nyc_school_tax_credit_fixed_amount_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for NYC School Tax Credit Fixed Amount"
    definition_period = YEAR
    defined_for = "in_nyc"

    def formula(tax_unit, period, parameters):
        # Eligibility is based on having a federal AGI below $30k
        # and being an NYC full-time resident.

        # Get the NYC School Tax Credit Fixed Amount part of the parameter tree.
        p = parameters(period).gov.local.ny.nyc.tax.income.credits.school.fixed

        # Get income that counts towards the NYC School Tax Credit.
        nyc_stc_income = tax_unit("nyc_school_credit_income", period)

        return nyc_stc_income <= p.income_limit



from policyengine_us.model_api import *


class hud_minimum_rent(Variable):
    value_type = float
    entity = SPMUnit
    label = "HUD minimum rent"
    unit = USD
    documentation = "Minimum Rent for HUD programs"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/cfr/text/24/5.630#a_3"

    def formula(spm_unit, period, parameters):
        # Assume all Section 8. Public housing is $50.
        return 25



from policyengine_us.model_api import *


class ky_modified_agi(Variable):
    value_type = float
    entity = TaxUnit
    label = "Kentucky modified adjusted gross income for the family size tax credit"
    unit = USD
    definition_period = YEAR
    reference = "https://revenue.ky.gov/Forms/740%20Packet%20Instructions%205-9-23.pdf#page=22"
    defined_for = StateCode.KY

    def formula(tax_unit, period, parameters):
        fed_agi = tax_unit("adjusted_gross_income", period)
        # Lump sum distributions which are not included in federal AGI are added to the federal AGI
        # Tax exempt interest from municipal bonds (non-Kentucky) is also added but excluded in this calculation
        tax_exempt_lump_sum = tax_unit(
            "form_4972_lumpsum_distributions", period
        )
        total_fed_agi = fed_agi + tax_exempt_lump_sum
        ky_agi = add(tax_unit, period, ["ky_agi"])
        # Lump sum distributions which are not included in federal AGI are added to the state AGI
        total_ky_agi = ky_agi + tax_exempt_lump_sum
        return max_(total_fed_agi, total_ky_agi)



from policyengine_us.model_api import *


class wi_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Wisconsin taxable income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.wi.gov/TaxForms2021/2021-Form1f.pdf"
        "https://www.revenue.wi.gov/TaxForms2021/2021-Form1-Inst.pdf"
        "https://www.revenue.wi.gov/TaxForms2022/2022-Form1f.pdf"
        "https://www.revenue.wi.gov/TaxForms2022/2022-Form1-Inst.pdf"
        "https://docs.legis.wisconsin.gov/misc/lfb/informational_papers/january_2023/0002_individual_income_tax_informational_paper_2.pdf"
    )
    defined_for = StateCode.WI

    def formula(tax_unit, period, parameters):
        taxinc = (
            tax_unit("wi_agi", period)
            - tax_unit("wi_standard_deduction", period)
            - tax_unit("wi_exemption", period)
        )
        return max_(0, taxinc)



from policyengine_us.model_api import *


class co_social_security_subtraction_indv_eligible(Variable):
    value_type = float
    entity = Person
    label = "Eligible for the Colorado social security subtraction for eligible individuals"
    defined_for = StateCode.CO
    unit = USD
    reference = (
        "https://tax.colorado.gov/sites/tax/files/documents/DR0104AD_2022.pdf#page=1",
        "https://tax.colorado.gov/sites/tax/files/documents/DR_104_Book_2022.pdf#page=12",
        "https://law.justia.com/codes/colorado/2022/title-39/article-22/part-1/section-39-22-104/",
        # C.R.S. 39-22-104(4)(g)(III)
    )
    definition_period = YEAR

    def formula(person, period, parameters):
        head = person("is_tax_unit_head", period)
        spouse = person("is_tax_unit_spouse", period)
        return head | spouse



from policyengine_us.model_api import *


class mn_additions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Minnesota additions to federal AGI"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.taxformfinder.org/forms/2021/2021-minnesota-form-m1m.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1m_22.pdf"
    )
    defined_for = StateCode.MN
    adds = "gov.states.mn.tax.income.additions.sources"



from policyengine_us.model_api import *


class long_term_capital_gains(Variable):
    value_type = float
    entity = Person
    label = "long-term capital gains"
    unit = USD
    documentation = "Net gains made from sales of assets held for more than one year (losses are expressed as negative gains)."
    definition_period = YEAR
    reference = dict(
        title="26 U.S. Code § 1222(3)",
        href="https://www.law.cornell.edu/uscode/text/26/1222#3",
    )
    uprating = "calibration.gov.irs.soi.long_term_capital_gains"
    adds = [
        "long_term_capital_gains_before_response",
        "capital_gains_behavioral_response",
    ]



from policyengine_us.model_api import *


class id_grocery_credit_qualified_months(Variable):
    value_type = int
    entity = Person
    label = "Months qualified for the Idaho grocery credit"
    definition_period = YEAR
    defined_for = StateCode.ID
    reference = (
        "https://law.justia.com/codes/idaho/2022/title-63/chapter-30/section-63-3024a/",
        "https://tax.idaho.gov/wp-content/uploads/forms/EFO00089/EFO00089_12-30-2022.pdf#page=7",
    )

    def formula(person, period, parameters):
        qualifying_months_sum = 0
        year = period.start.year
        for month in range(1, 13):
            monthly_period_str = f"{year}-{month:02d}"
            qualifying_months_sum += person(
                "id_grocery_credit_qualifying_month", monthly_period_str
            )
        return qualifying_months_sum



from policyengine_us.model_api import *


class de_relief_rebate(Variable):
    value_type = float
    entity = TaxUnit
    label = "Delaware relief rebate"
    unit = USD
    definition_period = YEAR
    reference = "https://legis.delaware.gov/BillDetail?LegislationId=99311"
    defined_for = StateCode.DE

    # Assuming that each adult in Delaware has filed a 2020
    # income tax return
    def formula(tax_unit, period, parameters):
        head_spouse_count = tax_unit("head_spouse_count", period)
        p = parameters(period).gov.states.de.tax.income.credits.relief_rebate
        return head_spouse_count * p.amount



from policyengine_us.model_api import *


class ca_state_disability_insurance(Variable):
    value_type = float
    entity = Person
    label = "California state disability insurance (SDI)"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.CA



from policyengine_us.model_api import *


class nj_agi_subtractions(Variable):
    value_type = float
    entity = Person
    label = "New Jersey subtractions from federal AGI by person"
    unit = USD
    documentation = "Subtractions from federal AGI to get NJ total income."
    definition_period = YEAR
    reference = "https://law.justia.com/codes/new-jersey/2022/title-54/section-54-8a-36/"
    defined_for = StateCode.NJ

    adds = "gov.states.nj.tax.income.subtractions"



from policyengine_us.model_api import *


class la_blind_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "Louisiana blind exemption"
    unit = USD
    definition_period = YEAR
    reference = [
        "https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf#page=1",
        "https://revenue.louisiana.gov/TaxForms/IT540iWEB(2022)D1.pdf#page=2",
    ]
    # Even though the tax computation worksheet refers "blind exemption" as credits, the instructions for
    # preparing tax form line 6a-6b specifies it as exemption. (the legal code also mentions deaf and totally disabled conditions,
    # but they are not mentioned either in the tax computation worksheet or tax form.)
    defined_for = StateCode.LA

    adds = ["la_blind_exemption_person"]



from policyengine_us.model_api import *


class ne_elderly_disabled_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "NE elderly/disabled tax credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://revenue.nebraska.gov/files/doc/tax-forms/2021/f_1040n_booklet.pdf"
        "https://revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_5.pdf"
    )
    defined_for = StateCode.NE

    def formula(tax_unit, period, parameters):
        credit = tax_unit("elderly_disabled_credit", period)
        us_itax_before_credits = tax_unit("income_tax_before_credits", period)
        return min_(credit, us_itax_before_credits)



from policyengine_us.model_api import *


class ca_care_poverty_line(Variable):
    value_type = float
    entity = Household
    definition_period = YEAR
    unit = USD
    label = "Poverty line as defined for California CARE program"
    reference = "https://leginfo.legislature.ca.gov/faces/codes_displaySection.xhtml?lawCode=PUC&sectionNum=739.1"
    defined_for = StateCode.CA

    def formula(household, period, parameters):
        n = household("household_size", period)
        # CARE treats one-person households as two-person households.
        adj_n = max_(n, 2)
        # Only need to look up Contiguous US poverty line.
        state_group = "CONTIGUOUS_US"
        p_fpg = parameters(period).gov.hhs.fpg
        p1 = p_fpg.first_person[state_group]
        pn = p_fpg.additional_person[state_group]
        return p1 + pn * (adj_n - 1)



from policyengine_us.model_api import *


class co_tanf_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Colorado TANF eligible"
    definition_period = YEAR
    defined_for = StateCode.CO

    def formula(spm_unit, period, parameters):
        # Assume federal demographic eligibility given consistency.
        demographic_eligible = spm_unit("is_demographic_tanf_eligible", period)
        income_eligible = spm_unit("co_tanf_income_eligible", period)
        return demographic_eligible & income_eligible



from policyengine_us.model_api import *


class capital_losses(Variable):
    value_type = float
    entity = Person
    label = "Capital losses (expressed as a non-negative number)"
    unit = USD
    documentation = "Losses from transactions involving property."
    definition_period = YEAR

    def formula(person, period, parameters):
        return max_(0, -person("capital_gains", period))



from policyengine_us.model_api import *


class co_chp(Variable):
    value_type = float
    entity = Person
    label = "Colorado Child Health Plan Plus expense savings"
    definition_period = YEAR
    defined_for = StateCode.CO

    def formula(person, period, parameters):
        saving = 0
        income_level = person("medicaid_income_level", period)
        is_pregnant = person("is_pregnant", period)
        p = parameters(period).gov.states.co.hcpf.chp
        elements = p.expenses
        for element in elements:
            copay = ~is_pregnant * p.copays[element].calc(income_level)
            expense = person(element + "_expense", period)
            saving += max_(0, expense - copay)
        return saving



from policyengine_us.model_api import *


class pell_grant_sai(Variable):
    value_type = float
    entity = Person
    unit = USD
    definition_period = YEAR
    label = "Pell Grant student aid index"

    def formula(person, period, parameters):
        head_contribution = person("pell_grant_head_contribution", period)
        dependent_contribution = person(
            "pell_grant_dependent_contribution", period
        )
        formula = person("pell_grant_formula", period)
        eligibility_type = person("pell_grant_eligibility_type", period)
        max_eligible = (
            eligibility_type == eligibility_type.possible_values.MAXIMUM
        )
        p = parameters(period).gov.ed.pell_grant.sai
        min_sai = p.limits.min_sai
        max_sai = where(max_eligible, 0, p.limits.max_sai)

        dependent_contribution_applies = formula == formula.possible_values.A
        applicable_dependent_contribution = (
            dependent_contribution_applies * dependent_contribution
        )
        unbound = head_contribution + applicable_dependent_contribution

        bound_ssi = min_(unbound, max_sai)
        return max_(bound_ssi, min_sai)



from policyengine_us.model_api import *


class md_total_subtractions(Variable):
    value_type = float
    entity = TaxUnit
    label = "MD total subtractions from AGI"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MD
    adds = "gov.states.md.tax.income.agi.subtractions.sources"



from policyengine_us.model_api import *


class nm_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Mexico taxable income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NM

    def formula(tax_unit, period, parameters):
        federal_agi = tax_unit("adjusted_gross_income", period)
        additions = tax_unit("nm_additions", period)
        salt_add_back = tax_unit("nm_salt_add_back", period)
        deductions = tax_unit("nm_deductions", period)
        exemptions = tax_unit("nm_exemptions", period)
        other_deductions_and_exemptions = tax_unit(
            "nm_other_deductions_and_exemptions", period
        )
        return max_(
            0,
            federal_agi
            + salt_add_back
            + additions
            - deductions
            - exemptions
            - other_deductions_and_exemptions,
        )



from policyengine_us.model_api import *


class ut_income_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "Utah income tax"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.UT
    reference = (
        "https://tax.utah.gov/forms/current/tc-40.pdf#page=2"  # line 38
    )

    adds = ["ut_income_tax_before_refundable_credits"]
    subtracts = ["ut_refundable_credits"]



from policyengine_us.model_api import *


class tax_unit_parents(Variable):
    value_type = int
    entity = TaxUnit
    label = "Number of parents in the tax unit"
    definition_period = YEAR

    adds = ["is_parent_of_filer_or_spouse"]



from policyengine_us.model_api import *


class az_military_retirement_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arizona military retirement subtraction"
    unit = USD
    reference = (
        "https://www.azleg.gov/ars/43/01022.htm",
        "https://azdor.gov/sites/default/files/2023-03/FORMS_INDIVIDUAL_2022_140BOOKLET.pdf#page=25",
        "https://azdor.gov/sites/default/files/2023-03/FORMS_INDIVIDUAL_2021_140BOOKLET.pdf#page=27",
    )
    definition_period = YEAR
    defined_for = StateCode.AZ

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.az.tax.income.subtractions.military_retirement
        person = tax_unit.members
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        military_retirement_pay = person("military_retirement_pay", period)
        eligible_military_pay = military_retirement_pay * head_or_spouse
        capped_military_pay = min_(eligible_military_pay, p.max_amount)
        return tax_unit.sum(capped_military_pay)



from policyengine_us.model_api import *


class la_receives_blind_exemption(Variable):
    value_type = bool
    entity = Person
    label = "Filer receives the Louisiana blind exemption over the subtraction"
    definition_period = YEAR
    #  RS 47:44.1 indicates that filers can only take either the exemption or subtraction.
    reference = "https://www.legis.la.gov/legis/Law.aspx?d=102133"
    defined_for = StateCode.LA
    default_value = True



from policyengine_us.model_api import *


class me_deduction_phaseout_percentage(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maine deduction phaseout percentage"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.mainelegislature.org/legis/statutes/36/title36sec5124-C.html",
        "https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_item_stand_%20ded_phaseout_wksht.pdf",
    )
    defined_for = StateCode.ME

    def formula(tax_unit, period, parameters):
        # Line 1. First get their Maine AGI.
        me_agi = tax_unit("me_agi", period)

        # Get filing status.
        filing_status = tax_unit("filing_status", period)

        # Get overall deduction part of parameters tree
        p = parameters(period).gov.states.me.tax.income.deductions.phase_out

        # Calculate the deduction phase-out parameters based on
        # filing status.
        phaseout_start = p.start[filing_status]
        excess = max_(me_agi - phaseout_start, 0)  # Line 3
        phaseout_width = p.width[filing_status]  # Line 4

        # Calculate the phaseout percent (Line 5)
        return min_(1, excess / phaseout_width)



from policyengine_us.model_api import *


class is_incarcerated(Variable):
    value_type = bool
    entity = Person
    label = "Is incarcerated"
    documentation = "Whether this person is incarcerated."
    definition_period = MONTH



from policyengine_us.model_api import *


class additional_senior_deduction_eligible_person(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    label = "Person is eligible for the additional senior deduction"
    reference = (
        "https://www.congress.gov/bill/119th-congress/house-bill/1/text"
    )

    def formula(person, period, parameters):
        ssn_card_type = person("ssn_card_type", period)
        ssn_card_str = ssn_card_type.decode_to_str()
        p = parameters(period).gov.irs.deductions.senior_deduction
        eligible_ssn_card_type = np.isin(
            ssn_card_str, p.eligible_ssn_card_type
        )
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        age = person("age", period)
        p_irs = parameters(period).gov.irs.deductions.standard.aged_or_blind
        aged = age >= p_irs.age_threshold
        return eligible_ssn_card_type & head_or_spouse & aged



from policyengine_us.model_api import *


class education_credit_phase_out(Variable):
    value_type = float
    entity = TaxUnit
    label = "Education credit phase-out"
    unit = "/1"
    documentation = "Percentage of the American Opportunity and Lifetime Learning credits which are phased out"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        education = parameters(period).gov.irs.credits.education
        agi = tax_unit("adjusted_gross_income", period)
        is_joint = tax_unit("tax_unit_is_joint", period)
        phase_out_start = where(
            is_joint,
            education.phase_out.start.joint,
            education.phase_out.start.single,
        )
        phase_out_length = where(
            is_joint,
            education.phase_out.length.joint,
            education.phase_out.length.single,
        )
        excess_agi = max_(0, agi - phase_out_start)
        return min_(1, excess_agi / phase_out_length)



from policyengine_us.model_api import *


class standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Standard deduction"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/63#c"
    adds = [
        "basic_standard_deduction",
        "additional_standard_deduction",
        "bonus_guaranteed_deduction",
    ]



from policyengine_us.model_api import *


class md_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "MD taxable income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MD

    def formula(tax_unit, period, parameters):
        md_agi = tax_unit("md_agi", period)
        deductions_exemptions = add(
            tax_unit, period, ["md_deductions", "md_exemptions"]
        )
        return max_(0, md_agi - deductions_exemptions)



from policyengine_us.model_api import *


class ssi(Variable):
    value_type = float
    entity = Person
    label = "SSI"
    documentation = "Supplemental Security Income"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/42/1382"

    def formula(person, period, parameters):
        if parameters(period).gov.ssa.ssi.abolish_ssi:
            return 0
        return max_(0, person("uncapped_ssi", period))



from policyengine_us.model_api import *


class net_capital_gain(Variable):
    value_type = float
    entity = TaxUnit
    label = "Net capital gain"
    unit = USD
    documentation = (
        "The excess of net long-term capital gain over net short-term capital"
        'loss, plus qualified dividends (the definition of "net capital gain"'
        "which applies to 26 U.S.C. § 1(h) from § 1(h)(11))."
    )
    definition_period = YEAR
    reference = dict(
        title="26 U.S. Code § 1222(11)",
        href="https://www.law.cornell.edu/uscode/text/26/1222#11",
    )

    def formula(tax_unit, period, parameters):
        lt_capital_gain = max_(
            0, add(tax_unit, period, ["long_term_capital_gains"])
        )
        st_capital_loss = max_(
            0, -add(tax_unit, period, ["short_term_capital_gains"])
        )
        net_cap_gain = max_(0, lt_capital_gain - st_capital_loss)
        qual_div_income = add(tax_unit, period, ["qualified_dividend_income"])
        return net_cap_gain + qual_div_income



from policyengine_us.model_api import *
from policyengine_core.simulations import Simulation
from policyengine_us.variables.household.demographic.geographic.ucgid.ucgid_enum import (
    UCGID,
)
import numpy as np


class ucgid(Variable):
    value_type = Enum
    possible_values = UCGID
    default_value = UCGID.US
    entity = Household
    label = "Unified Congressional Geographic Identifier (UCGID)"
    definition_period = YEAR
    documentation = """
    Unified Congressional Geographic Identifier (UCGID) for the household as defined by the U.S. Census Bureau.
    """

    def formula(household, period, parameters):
        simulation: Simulation = household.simulation

        # Check if this variable has been set as input
        if simulation.get_holder("ucgid").get_known_periods():
            return household("ucgid", period)

        # Try to derive from state_code
        state_code_enum = household("state_code", period)
        state_code_strings = state_code_enum.decode_to_str()

        # Map state codes to UCGID state values
        state_mapping = {
            "AL": UCGID.AL,
            "AK": UCGID.AK,
            "AZ": UCGID.AZ,
            "AR": UCGID.AR,
            "CA": UCGID.CA,
            "CO": UCGID.CO,
            "CT": UCGID.CT,
            "DE": UCGID.DE,
            "DC": UCGID.DC,
            "FL": UCGID.FL,
            "GA": UCGID.GA,
            "HI": UCGID.HI,
            "ID": UCGID.ID,
            "IL": UCGID.IL,
            "IN": UCGID.IN,
            "IA": UCGID.IA,
            "KS": UCGID.KS,
            "KY": UCGID.KY,
            "LA": UCGID.LA,
            "ME": UCGID.ME,
            "MD": UCGID.MD,
            "MA": UCGID.MA,
            "MI": UCGID.MI,
            "MN": UCGID.MN,
            "MS": UCGID.MS,
            "MO": UCGID.MO,
            "MT": UCGID.MT,
            "NE": UCGID.NE,
            "NV": UCGID.NV,
            "NH": UCGID.NH,
            "NJ": UCGID.NJ,
            "NM": UCGID.NM,
            "NY": UCGID.NY,
            "NC": UCGID.NC,
            "ND": UCGID.ND,
            "OH": UCGID.OH,
            "OK": UCGID.OK,
            "OR": UCGID.OR,
            "PA": UCGID.PA,
            "RI": UCGID.RI,
            "SC": UCGID.SC,
            "SD": UCGID.SD,
            "TN": UCGID.TN,
            "TX": UCGID.TX,
            "UT": UCGID.UT,
            "VT": UCGID.VT,
            "VA": UCGID.VA,
            "WA": UCGID.WA,
            "WV": UCGID.WV,
            "WI": UCGID.WI,
            "WY": UCGID.WY,
        }

        # Convert state code strings to UCGID values
        result = np.empty(len(state_code_strings), dtype=object)
        for i, state_str in enumerate(state_code_strings):
            if state_str in state_mapping:
                result[i] = state_mapping[state_str]
            else:
                result[i] = UCGID.US

        return result



from policyengine_us.model_api import *


class va_subtractions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Virginia subtractions from the adjusted gross income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.VA

    adds = "gov.states.va.tax.income.subtractions.subtractions"



from policyengine_us.model_api import *


class mi_retirement_benefits_deduction_tier_three_ss_exempt_retired(Variable):
    value_type = float
    entity = TaxUnit
    label = "Michigan retired tier three retirement benefits deduction"
    unit = USD
    definition_period = YEAR
    reference = (
        "http://legislature.mi.gov/doc.aspx?mcl-206-30",  # (9)(c)
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf#page=21",
        "https://www.michigan.gov/taxes/iit/retirement-and-pension-benefits",
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/4884.pdf",
    )
    defined_for = "mi_retirement_benefits_deduction_tier_three_eligible"

    def formula(tax_unit, period, parameters):
        # Modeled after 2022 Michigan Pension Schedule (Form 4884) Section B
        p = parameters(
            period
        ).gov.states.mi.tax.income.deductions.retirement_benefits.tier_three.ss_exempt.retired
        # Recipients should received retirement benefits from SSA exempt employment
        # and were retired before qualifying year
        ss_retired_eligible_people = tax_unit(
            "mi_retirement_benefits_deduction_tier_three_ss_exempt_retired_eligible_people",
            period,
        )

        filing_status = tax_unit("filing_status", period)
        person = tax_unit.members
        uncapped_pension_income = person("taxable_pension_income", period)
        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)

        # Where one or two people in the household qualify determines the amount of deduction
        # Line 11 & 12
        cap = select(
            [ss_retired_eligible_people == 1, ss_retired_eligible_people > 1],
            [
                p.single_qualifying_amount[filing_status],
                p.both_qualifying_amount[filing_status],
            ],
            default=0,
        )
        uncapped_head_or_spouse_pension = tax_unit.sum(
            uncapped_pension_income * is_head_or_spouse
        )

        # Modeled after Worksheet 3 for Filers with Taxable Railroad Retirement Benefits or
        # Qualifying Retirement and Pension Benefits from Service in the U.S. Armed Forces or
        # Michigan National Guard and Completing Section B of Form 4884

        # If a filer is recieving military retirement pay, the calculation includes the smaller of
        # the tier one or tier three deduction amount
        military_retirement_pay_eligible = (
            tax_unit.sum(person("military_retirement_pay", period)) > 0
        )
        # Line 8
        tier_one_amount = tax_unit(
            "mi_retirement_benefits_deduction_tier_one_amount",
            period,
        )
        # Line 10
        smaller_of_cap_or_tier_one_amount = min_(cap, tier_one_amount)

        eligible_deduction = where(
            military_retirement_pay_eligible,
            smaller_of_cap_or_tier_one_amount,
            cap,
        )
        # Line 17
        return min_(uncapped_head_or_spouse_pension, eligible_deduction)



from policyengine_us.model_api import *


class is_full_time_student(Variable):
    value_type = bool
    entity = Person
    label = "Is a full time student"
    definition_period = YEAR

    def formula(person, period, parameters):
        in_college = person("is_full_time_college_student", period)
        in_k12 = person("is_in_k12_school", period)
        return in_college | in_k12



from policyengine_us.model_api import *


class vt_csrs_retirement_pay_exclusion(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Vermont Civil Service Retirement System (CSRS) retirement income exclusion"
    reference = (
        "https://tax.vermont.gov/individuals/seniors-and-retirees",
        "https://legislature.vermont.gov/statutes/section/32/151/05830e",  # Legal Code Titl. 32 V.S.A. § 5830e (b)(1)(B), (b)(2)(B)
    )
    unit = USD
    defined_for = StateCode.VT
    documentation = "Vermont Civil Service Retirement System (CSRS) retirement benefits exempt from Vermont taxation."

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        p = parameters(
            period
        ).gov.states.vt.tax.income.agi.retirement_income_exemption.csrs
        # Get retirement amount from military retirement system
        tax_unit_csrs_retirement_pay = add(
            tax_unit, period, ["csrs_retirement_pay"]
        )
        # Retirement income from systems other than social security have maximum amount.
        return min_(tax_unit_csrs_retirement_pay, p.amount)



from policyengine_us.model_api import *


class ma_part_a_taxable_capital_gains_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "MA Part A taxable income from short-term capital gains"
    unit = USD
    definition_period = YEAR
    reference = "https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions"
    defined_for = StateCode.MA

    def formula(tax_unit, period, parameters):
        part_a_agi = tax_unit("ma_part_a_agi", period)
        dividends = add(tax_unit, period, ["dividend_income"])
        stcg_agi = part_a_agi - dividends
        div_excess_exemption = tax_unit(
            "ma_part_a_div_excess_exemption", period
        )
        return max_(0, stcg_agi - div_excess_exemption)



from policyengine_us.model_api import *


class co_ctc_eligible_children_count(Variable):
    value_type = int
    entity = TaxUnit
    label = "Colorado child tax credit eligible children count"
    definition_period = YEAR
    reference = (
        "https://leg.colorado.gov/sites/default/files/2023a_1112_signed.pdf#page=2",
    )
    defined_for = StateCode.CO

    adds = ["co_ctc_eligible_child"]



from policyengine_us.model_api import *


class wi_itemized_deduction_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Wisconsin itemized deduction credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.wi.gov/TaxForms2021/2021-Form1f.pdf#page=4"
        "https://www.revenue.wi.gov/TaxForms2022/2022-Form1f.pdf#page=4"
        "https://docs.legis.wisconsin.gov/misc/lfb/informational_papers/january_2023/0002_individual_income_tax_informational_paper_2.pdf#page=19"
    )
    defined_for = StateCode.WI

    def formula(tax_unit, period, parameters):
        wi_itmded = tax_unit("itemized_deductions_less_salt", period)
        wi_stdded = tax_unit("wi_standard_deduction", period)
        excess_itmded = max_(0, wi_itmded - wi_stdded)
        p = parameters(period).gov.states.wi.tax.income
        return excess_itmded * p.credits.itemized_deduction.rate



from policyengine_us.model_api import *


class is_in_secondary_school(Variable):
    value_type = bool
    entity = Person
    label = "Is in secondary school (or in an equivalent level of training)"
    definition_period = YEAR



from policyengine_us.model_api import *


class mn_subtractions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Minnesota subtractions from federal AGI"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.state.mn.us/sites/default/files/2021-12/m1_21_0.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1_inst_21.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2022-12/m1_22.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-03/m1_inst_22.pdf"
    )
    defined_for = StateCode.MN
    adds = "gov.states.mn.tax.income.subtractions.sources"



from policyengine_us.model_api import *


class ma_tafdc_clothing_allowance(Variable):
    value_type = float
    unit = USD
    entity = Person
    label = "Massachusetts Temporary Assistance for Families with Dependent Children (TAFDC) clothing allowance"
    definition_period = YEAR
    reference = "https://www.masslegalservices.org/content/75-how-much-will-you-get-each-month"
    defined_for = StateCode.MA

    def formula(person, period, parameters):
        eligible_child = person("ma_tafdc_eligible_dependent", period)
        p = parameters(period).gov.states.ma.dta.tcap.tafdc
        return eligible_child * p.clothing_allowance



from policyengine_us.model_api import *


class spm_unit_total_income_reported(Variable):
    value_type = float
    entity = SPMUnit
    label = "SPM unit total income"
    definition_period = YEAR
    unit = USD



from policyengine_us.model_api import *


class divorce_year(Variable):
    value_type = int
    entity = Person
    label = "The year that the person was divorced."
    definition_period = YEAR
    default_value = 2010



from policyengine_us.model_api import *


class la_non_refundable_cdcc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Louisiana non-refundable Child and Dependent Care Credit"
    unit = USD
    definition_period = YEAR
    reference = "http://legis.la.gov/Legis/Law.aspx?d=101769"
    defined_for = StateCode.LA

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.la.tax.income.credits.cdcc.non_refundable
        # Louisiana matches the potential federal credit
        us_cdcc = tax_unit("cdcc_potential", period)
        us_agi = tax_unit("adjusted_gross_income", period)
        match = p.match.calc(us_agi, right=True)
        uncapped_credit = us_cdcc * match
        # The non refundable credit is capped for taxpayers with AGI above the threshold
        top_threshold = p.match.thresholds[-1]
        agi_over_cap_threshold = us_agi > top_threshold
        capped_credit = min_(uncapped_credit, p.upper_bracket_cap)
        return where(agi_over_cap_threshold, capped_credit, uncapped_credit)



from policyengine_us.model_api import *


class is_snap_ineligible_student(Variable):
    value_type = bool
    entity = Person
    label = "Is an ineligible student for SNAP"
    documentation = "Whether this person is an ineligible student for SNAP and can not be counted towards the household size"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/7/2015"

    def formula(person, period, parameters):
        student = person(
            "is_full_time_student", period
        )  # part-time students should also count
        age = person("age", period)
        p = parameters(period).gov.usda.snap.student
        age_eligible = p.age_threshold.calc(age)
        disabled = person("is_disabled", period)
        hours_worked = person("weekly_hours_worked", period)
        hours_eligible = hours_worked >= p.working_hours_threshold
        # The parental status applies to children under 6 (under 12
        # if no care available)
        # Single parents with children under 12 are also eligible
        is_parent = person("is_parent", period)
        return student & ~(
            age_eligible | disabled | hours_eligible | is_parent
        )



from policyengine_us.model_api import *


class va_must_file(Variable):
    value_type = float
    entity = TaxUnit
    label = "Tax unit must file Virginia income taxes"
    unit = USD
    definition_period = YEAR
    reference = "https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2023-760-instructions.pdf#page=10"
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        agi = tax_unit("va_agi", period)
        p = parameters(period).gov.states.va.tax.income
        filing_status = tax_unit("filing_status", period)
        return agi >= p.filing_requirement[filing_status]



from policyengine_us.model_api import *


class ia_itemized_deductions_unit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Iowa itemized deductions for tax unit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.iowa.gov/sites/default/files/2022-01/IA1040%2841-001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-01/2021%20Expanded%20Instructions_010323.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-01/2022IA1040%2841001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-03/2022%20Expanded%20Instructions_022023.pdf"
    )
    defined_for = StateCode.IA

    def formula(tax_unit, period, parameters):
        """
        FROM THE 2021 INSTRUCTIONS (PAGE 46):
          The $10,000 federal cap on the itemized deduction for state and
        local taxes does not apply for Iowa purposes. Taxpayers may still
        deduct eligible state and local taxes paid, independent of the
        federal dollar limitation.
        """
        # compute tax unit's itemized deductions
        p = parameters(period).gov.states.ia.tax.income.deductions.itemized
        if p.applies_federal:
            p_fed = parameters(period).gov.irs.deductions
            return add(tax_unit, period, p_fed.itemized_deductions)
        else:
            return add(
                tax_unit,
                period,
                ["itemized_deductions_less_salt", "real_estate_taxes"],
            )



from policyengine_us.model_api import *


class wv_public_pension_subtraction_person(Variable):
    value_type = float
    entity = Person
    label = "West Virginia public pension subtraction for each person"
    defined_for = StateCode.WV
    unit = USD
    definition_period = YEAR
    reference = "https://code.wvlegislature.gov/11-21-12/"

    def formula(person, period, parameters):
        # "benefits received under any federal retirement system to which Title 4 U.S.C. §111 applies"
        # https://www.law.cornell.edu/uscode/text/4/111
        federal_pension_income = person(
            "taxable_federal_pension_income", period
        )
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        p = parameters(
            period
        ).gov.states.wv.tax.income.subtractions.public_pension
        capped = min_(federal_pension_income, p.max_amount)
        return capped * head_or_spouse



from policyengine_us.model_api import *


class ne_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "NE standard deduction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://revenue.nebraska.gov/files/doc/tax-forms/2021/f_1040n_booklet.pdf"
        "https://revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_5.pdf"
    )
    defined_for = StateCode.NE

    def formula(tax_unit, period, parameters):
        base_state_sd = tax_unit("ne_base_standard_deduction", period)
        base_fed_sd = tax_unit("basic_standard_deduction", period)
        smaller_base_sd = min_(base_state_sd, base_fed_sd)
        additional_sd = tax_unit("additional_standard_deduction", period)
        return smaller_base_sd + additional_sd



from policyengine_us.model_api import *


class used_clean_vehicle_sale_price(Variable):
    value_type = float
    entity = TaxUnit
    unit = USD
    label = "Sale price of newly purchased used clean vehicle"
    definition_period = YEAR
    quantity_type = STOCK



from policyengine_us.model_api import *


class hi_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Hawaii exemptions"
    unit = USD
    documentation = (
        "https://files.hawaii.gov/tax/forms/2022/n11ins.pdf#page=20"
    )
    definition_period = YEAR
    defined_for = StateCode.HI

    def formula(tax_unit, period, parameters):
        regular_exemptions = tax_unit("hi_regular_exemptions", period)
        disabled_exemptions = tax_unit("hi_disabled_exemptions", period)
        return max_(regular_exemptions, disabled_exemptions)



from policyengine_us.model_api import *


class ms_regular_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "Mississippi regular exemption"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MS

    def formula(tax_unit, period, parameters):
        # First get their filing status.
        filing_status = tax_unit("filing_status", period)

        # Then get the MS Exemptions part of the parameter tree.
        p = parameters(period).gov.states.ms.tax.income.exemptions

        # Get their regular exemption amount based on their filing status.
        return p.regular.amount[filing_status]



from policyengine_us.model_api import *


class va_non_refundable_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Virginia non-refundable EITC"
    unit = USD
    definition_period = YEAR
    reference = "https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf#page=32"
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        # Either claims refundable or non-refundable, but not both.
        claims = ~tax_unit("va_claims_refundable_eitc", period)
        amount_if_claimed = tax_unit(
            "va_non_refundable_eitc_if_claimed", period
        )
        return claims * amount_if_claimed



from policyengine_us.model_api import *


class ca_investment_interest_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "California investment interest deduction"
    unit = USD
    reference = (
        "https://www.ftb.ca.gov/forms/2021/2021-3526.pdf"
        "https://law.justia.com/codes/california/2022/code-rtc/division-2/part-11/chapter-7/article-1/section-24344/"
    )
    definition_period = YEAR
    defined_for = StateCode.CA

    def formula(tax_unit, period, parameters):
        # Line 1
        investment_interest_expense = add(
            tax_unit, period, ["investment_interest_expense"]
        )
        # Line 5
        investment_expenses = add(tax_unit, period, ["investment_expenses"])
        # Line 6
        net_investment_income = (
            investment_interest_expense - investment_expenses
        )
        # Line 8
        investment_interest_expense_deduction = min_(
            investment_interest_expense, net_investment_income
        )
        # Line 9
        form_4952_amount = tax_unit("investment_income_form_4952", period)
        # Line 10
        return np.abs(form_4952_amount - investment_interest_expense_deduction)



from policyengine_us.model_api import *


class or_income_subtractions(Variable):
    value_type = float
    entity = TaxUnit
    label = "OR income subtractions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.oregon.gov/dor/forms/FormsPubs/form-or-40-inst_101-040-1_2021.pdf#page=13",
        "https://www.oregonlegislature.gov/bills_laws/ors/ors316.html",  # Subsection 316.800
    )
    defined_for = StateCode.OR

    adds = "gov.states.or.tax.income.subtractions.subtractions"



from policyengine_us.model_api import *


class cdcc_limit(Variable):
    value_type = float
    entity = TaxUnit
    label = "CDCC-relevant care expense limit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/21#c"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.irs.credits.cdcc
        capped_count_cdcc_eligible = tax_unit(
            "capped_count_cdcc_eligible", period
        )
        return p.max * capped_count_cdcc_eligible



from policyengine_us.model_api import *


class snap_earned_income_person(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = MONTH
    label = "SNAP earned income person"
    documentation = "Earned income per person for calculating the SNAP earned income deduction"
    reference = "https://www.law.cornell.edu/cfr/text/7/273.9#b_1"
    unit = USD
    defined_for = "snap_countable_earner"

    adds = "gov.usda.snap.income.sources.earned"



from policyengine_us.model_api import *


class nc_scca_fpg_rate(Variable):
    value_type = float
    entity = SPMUnit
    label = "North Carolina Subsidized Child Care Assistance (SCCA) program income limits compared to the FPL"
    reference = "https://policies.ncdhhs.gov/wp-content/uploads/chapter-8-parental-fees-7.pdf#page=8"
    definition_period = YEAR
    defined_for = StateCode.NC

    def formula(spm_unit, period, parameters):
        # Entry income eligibility depends on the child's age and disability status:
        # - Children aged 0-5 (preschool): 200% FPL
        # - Children aged 6-12 (school age): 133% FPL
        # - Children with special needs (any age): 200% FPL
        p = parameters(period).gov.states.nc.ncdhhs.scca

        # Retrieve age, school age status, and disability status for all members
        person = spm_unit.members
        is_school_age = person("nc_scca_is_school_age", period)
        disabled = person("is_disabled", period)

        # Check if any child is disabled and age-eligible for the program
        has_eligible_disabled_child = spm_unit.any(
            person("nc_scca_child_age_eligible", period) & disabled
        )

        # Check if all children are school age (if there are any children)
        has_any_children = spm_unit.any(person("is_child", period))
        all_children_school_age = ~has_any_children | (
            has_any_children
            & spm_unit.all(~person("is_child", period) | is_school_age)
        )

        # If there are any non-school age children or disabled children,
        # use the higher non-school age FPG limit, otherwise use school age limit
        has_preschool_or_special_needs = (
            has_eligible_disabled_child | ~all_children_school_age
        )

        # Select the appropriate FPG limit based on household composition
        # Children under school age (defined in p.age.school) or with special needs: 200% FPL
        # Only school-age children without special needs: 133% FPL
        return where(
            has_preschool_or_special_needs,
            p.entry.fpg_limit.preschool,
            p.entry.fpg_limit.school_age,
        )



from policyengine_us.model_api import *


class substitution_elasticity_lsr(Variable):
    value_type = float
    entity = Person
    label = "substitution elasticity of labor supply response"
    unit = "/1"
    definition_period = YEAR
    requires_computation_after = "labor_supply_behavioral_response"

    def formula(person, period, parameters):
        raw_earnings = add(
            person,
            period,
            [
                "employment_income_before_lsr",
                "self_employment_income_before_lsr",
            ],
        )
        earnings = max_(raw_earnings, 0)
        wage_change = person("relative_wage_change", period)
        return (
            earnings * wage_change * person("substitution_elasticity", period)
        )



from policyengine_us.model_api import *


class ca_tanf(Variable):
    value_type = float
    entity = SPMUnit
    label = "California CalWORKs Cash Benefit"
    unit = USD
    definition_period = YEAR
    defined_for = "ca_tanf_eligible"

    def formula(spm_unit, period, parameters):
        maximum_payment = spm_unit("ca_tanf_maximum_payment", period)
        countable_income = spm_unit(
            "ca_tanf_countable_income_recipient", period
        )
        person = spm_unit.members
        eligible_people_based_on_immigration_status = spm_unit.sum(
            person("ca_tanf_immigration_status_eligible_person", period)
        )
        spm_unit_size = spm_unit("spm_unit_size", period)
        prorated_fraction = np.zeros_like(spm_unit_size, dtype=float)
        mask = spm_unit_size != 0
        prorated_fraction[mask] = (
            eligible_people_based_on_immigration_status[mask]
            / spm_unit_size[mask]
        )
        return prorated_fraction * max_(maximum_payment - countable_income, 0)



from policyengine_us.model_api import *


class hi_cdcc_min_head_spouse_earned(Variable):
    value_type = float
    entity = TaxUnit
    label = "Hawaii minimum income between head and spouse for the CDCC"
    defined_for = StateCode.HI
    unit = USD
    definition_period = YEAR
    reference = (
        "https://files.hawaii.gov/tax/forms/2022/n11ins.pdf#page=28"
        "https://files.hawaii.gov/tax/forms/2022/n11ins.pdf#page=29"
        "https://files.hawaii.gov/tax/legal/hrs/hrs_235.pdf#page=41"
        "https://files.hawaii.gov/tax/forms/2022/schx_i.pdf#page=2"
    )

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.hi.tax.income.credits.cdcc
        person = tax_unit.members
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        earnings = person("earned_income", period)
        # Head or spouse are eligible for an income floor if disabled or a student
        # Start with case with only one floor-eligible spouse.
        qualified_children = person.tax_unit("count_cdcc_eligible", period)
        # Floor depends on number of eligible dependents
        floor = p.disabled_student_income_floor.calc(qualified_children)
        floor_eligible = person("hi_cdcc_income_floor_eligible", period)
        potential_floored_earnings = max_(floor, earnings)
        # floored_earnings = where(floor_eligible, floored_earnings, earnings)
        floored_earnings = where(
            floor_eligible, potential_floored_earnings, earnings
        )
        # To take the lesser of head or spouse floored earnings,
        # assign infinite earnings to dependents.
        lesser_floored_earnings = tax_unit.min(
            where(head_or_spouse, floored_earnings, np.inf)
        )
        # Case with two floor-eligible spouses:
        # If both spouses are students or disabled, only one can claim the floor.
        # To maximize the credit value, since the credit phases in with respect to (floored) earnings,
        # the filer should floor the spouse with the smaller earnings.
        # Note that the phase-out is separate and with respect to AGI.
        # Ignore the floor, just take the greater unfloored earnings.
        # This is equivalent to only applying the floor to one spouse (that with lesser earnings).
        head_or_spouse_earning = head_or_spouse * earnings
        greater_earnings = tax_unit.max(head_or_spouse_earning)
        # Select based on number of floored.
        count_floor_eligible = tax_unit.sum(head_or_spouse * floor_eligible)
        not_both_eligible_above_floor = (count_floor_eligible == 2) & (
            tax_unit.sum(head_or_spouse_earning > floor) != 2
        )

        return where(
            not_both_eligible_above_floor,
            greater_earnings,
            lesser_floored_earnings,
        )



from policyengine_us.model_api import *
from policyengine_us.variables.gov.states.il.tax.income.exemptions.il_personal_exemption_eligibility_status import (
    ILPersonalExemptionEligibilityStatus,
)


class il_personal_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "Illinois personal exemption amount"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.IL

    def formula(tax_unit, period, parameters):
        il_is_personal_exemption_eligible = tax_unit(
            "il_personal_exemption_eligibility_status", period
        )

        personal_exemption_amount = parameters(
            period
        ).gov.states.il.tax.income.exemption.personal

        return personal_exemption_amount * select(
            [
                il_is_personal_exemption_eligible
                == ILPersonalExemptionEligibilityStatus.BOTH_ELIGIBLE,
                il_is_personal_exemption_eligible
                == ILPersonalExemptionEligibilityStatus.PARTIALLY_ELIGIBLE,
            ],
            [2, 1],
            0,
        )



from policyengine_us.model_api import *


class in_other_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Indiana other deductions"
    unit = USD
    definition_period = YEAR
    documentation = "Other deductions available in Indiana including those for civil service annuities, disability retirement, government or civic group capital contributions, human services for Medicaid recipients,  infrastructure fund gifts, indiana lottery winings annuity, Indiana partnership long-term care policy premiums, military retirement income or survivor's benefits, National Guard and reserve component members, Olympic/Paralymic medal winners, qualified patents income eemption, railroad unemployment and sickness benefits, repayment of previously taxed income deductions, COVID-related employee retention credit dissalowed expenses, and Indiana-only tax-exempt bonds."
    reference = "http://iga.in.gov/legislative/laws/2021/ic/titles/006#6-3-2"
    defined_for = StateCode.IN



from policyengine_us.model_api import *


class il_tanf_income_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for Illinois Temporary Assistance for Needy Families (TANF) due to income"
    definition_period = MONTH
    reference = "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-112.155"
    defined_for = StateCode.IL

    def formula(spm_unit, period, parameters):
        countable_income = spm_unit(
            "il_tanf_countable_income_for_initial_eligibility", period
        )
        payment_level = spm_unit(
            "il_tanf_payment_level_for_initial_eligibility", period
        )
        return countable_income <= payment_level



from policyengine_us.model_api import *


class taxable_social_security(Variable):
    value_type = float
    entity = Person
    label = "Taxable Social Security"
    unit = USD
    definition_period = YEAR

    def formula(person, period, parameters):
        unit_tss = person.tax_unit("tax_unit_taxable_social_security", period)
        # allocate unit_tss to head and spouse in proportion to social_security
        ind_socsec = person("social_security", period)
        unit_socsec = person.tax_unit("tax_unit_social_security", period)
        # Compute each person's share of tax unit's Social Security benefits.
        # Use mask instead of where to avoid divide-by-zero warnings. Default to zero.
        person_frac = np.zeros_like(unit_socsec)
        mask = unit_socsec > 0
        person_frac[mask] = ind_socsec[mask] / unit_socsec[mask]
        is_spouse = person("is_tax_unit_spouse", period)
        spouse_frac = min_(1, where(is_spouse, person_frac, 0))
        unit_spouse_frac = person.tax_unit.sum(spouse_frac)
        is_head = person("is_tax_unit_head", period)
        head_frac = where(is_head, 1 - unit_spouse_frac, 0)
        unit_head_frac = person.tax_unit.sum(head_frac)
        return unit_tss * select(
            [is_head, is_spouse], [unit_head_frac, unit_spouse_frac], default=0
        )



from policyengine_us.model_api import *
from policyengine_core.simulations import Simulation


class slcsp_age_0(Variable):
    value_type = float
    entity = Household
    label = "Second-lowest ACA silver-plan for a person aged 0"
    unit = USD
    definition_period = MONTH

    def formula(household, period, parameters):
        # Get state code and rating area
        state_code = household("state_code", period)
        rating_area = household("slcsp_rating_area", period)

        # Access the baseline costs from parameters
        p = parameters(period).gov.aca

        # Return the cost for the specific state and rating area
        return p.state_rating_area_cost[state_code][rating_area]



from policyengine_us.model_api import *


class oh_joint_filing_credit_qualifying_income(Variable):
    value_type = float
    entity = Person
    label = "Ohio qualifying income for the joint filing credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://codes.ohio.gov/ohio-revised-code/section-5747.055",
        "https://tax.ohio.gov/static/forms/ohio_individual/individual/2021/pit-it1040-booklet.pdf#page=20",
    )
    defined_for = StateCode.OH

    adds = ["oh_agi_person"]

    subtracts = ["oh_joint_filing_credit_agi_subtractions"]



from policyengine_us.model_api import *


class ca_amti_adjustments(Variable):
    value_type = float
    entity = TaxUnit
    label = "California AMTI adjustment"
    defined_for = StateCode.CA
    unit = USD
    definition_period = YEAR
    reference = "https://www.ftb.ca.gov/forms/2022/2022-540-p.pdf"

    def formula(tax_unit, period, parameters):
        # Line 1 -14
        p = parameters(period).gov.states.ca.tax.income.amt
        itemized_ded = tax_unit("ca_itemized_deductions", period)
        standard_ded = tax_unit("ca_standard_deduction", period)
        # Assuming that tax filer claimed itemized deducions if above standard ded amount
        itemizes = itemized_ded > standard_ded
        itemized_sources = add(tax_unit, period, p.amti.sources)
        # Line 6
        deductions = where(itemizes, itemized_sources, standard_ded)
        # Line 7
        investment_ded = tax_unit("ca_investment_interest_deduction", period)
        # line 14
        return investment_ded + deductions



from policyengine_us.model_api import *


class ks_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Kansas taxable income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.ksrevenue.gov/pdf/ip21.pdf"
        "https://www.ksrevenue.gov/pdf/ip22.pdf"
    )
    defined_for = StateCode.KS

    def formula(tax_unit, period, parameters):
        std_ded = tax_unit("ks_standard_deduction", period)
        itm_ded = tax_unit("ks_itemized_deductions", period)
        deductions = where(itm_ded > std_ded, itm_ded, std_ded)
        exemptions = tax_unit("ks_exemptions", period)
        return max_(0, tax_unit("ks_agi", period) - deductions - exemptions)



from policyengine_us.model_api import *
from numpy import ceil


class mn_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Minnesota exemptions amount"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.state.mn.us/sites/default/files/2021-12/m1_21_0.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1_inst_21.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2022-12/m1_22.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-03/m1_inst_22.pdf"
    )
    defined_for = StateCode.MN

    def formula(tax_unit, period, parameters):
        dependents = tax_unit("tax_unit_dependents", period)
        p = parameters(period).gov.states.mn.tax.income.exemptions
        exemptions = p.amount * dependents
        # limit exemptions if federal AGI is above a threshold
        agi = tax_unit("adjusted_gross_income", period)
        filing_status = tax_unit("filing_status", period)
        excess_agi = max_(0, agi - p.agi_threshold[filing_status])
        steps = ceil(excess_agi / p.agi_step_size[filing_status])
        offset_fraction = p.agi_step_fraction * steps
        offset = offset_fraction * exemptions
        return max_(0, exemptions - offset)



from policyengine_us.model_api import *


def create_tax_exempt() -> Reform:
    class irs_gross_income(Variable):
        value_type = float
        entity = Person
        label = "Gross income"
        unit = USD
        documentation = (
            "Gross income, as defined in the Internal Revenue Code."
        )
        definition_period = YEAR
        reference = "https://www.law.cornell.edu/uscode/text/26/61"

        def formula(person, period, parameters):
            sources = parameters(period).gov.irs.gross_income.sources
            total = 0
            not_dependent = ~person("is_tax_unit_dependent", period)
            for source in sources:
                # Add positive values only - losses are deducted later.
                total += not_dependent * max_(0, add(person, period, [source]))
            exempt_income = 0
            p = parameters(period).gov.contrib.tax_exempt
            if p.tip_income.income_tax_exempt:
                tip_income = person("tip_income", period)
                exempt_income += tip_income
            if p.overtime.income_tax_exempt:
                exempt_income += person("fsla_overtime_premium", period)
            return max_(total - exempt_income, 0)

    class payroll_tax_gross_wages(Variable):
        value_type = float
        entity = Person
        label = "Gross wages and salaries for payroll taxes"
        definition_period = YEAR
        unit = USD

        def formula(person, period, parameters):
            income = person("irs_employment_income", period)
            p = parameters(period).gov.contrib.tax_exempt
            exempt_income = 0
            if p.tip_income.payroll_tax_exempt:
                exempt_income += person("tip_income", period)
            if p.overtime.payroll_tax_exempt:
                exempt_income += person("fsla_overtime_premium", period)
            return max_(income - exempt_income, 0)

    class reform(Reform):
        def apply(self):
            self.update_variable(irs_gross_income)
            self.update_variable(payroll_tax_gross_wages)

    return reform


def create_tax_exempt_reform(parameters, period, bypass: bool = False):
    if bypass:
        return create_tax_exempt()

    p = parameters(period).gov.contrib.tax_exempt

    if p.in_effect:
        return create_tax_exempt()
    else:
        return None


tax_exempt_reform = create_tax_exempt_reform(None, None, bypass=True)



from policyengine_us.model_api import *


class taxable_income_deductions_if_not_itemizing(Variable):
    value_type = float
    entity = TaxUnit
    label = "Deductions if not itemizing"
    unit = USD
    definition_period = YEAR

    adds = "gov.irs.deductions.deductions_if_not_itemizing"



from policyengine_us.model_api import *


class oh_personal_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Ohio personal exemptions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.ohio.gov/static/forms/ohio_individual/individual/2021/pit-it1040-booklet.pdf#page=14",
    )
    defined_for = StateCode.OH

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.oh.tax.income.exemptions.personal

        eligible_exemptions = add(
            tax_unit, period, ["oh_personal_exemptions_eligible_person"]
        )

        agi = tax_unit("oh_agi", period)
        exemption_amount = p.amount.calc(agi, right=True)

        return eligible_exemptions * exemption_amount



from policyengine_us.model_api import *


class mo_pension_and_ss_or_ssd_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Missouri Pension and Social Security or SS Disability Deduction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.mo.gov/forms/MO-A_2021.pdf#page=3",
        "https://dor.mo.gov/forms/MO-1040%20Fillable%20Calculating_2021.pdf#page=2",
        "https://revisor.mo.gov/main/OneSection.aspx?section=143.124",
    )
    defined_for = StateCode.MO

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        # Section C, SS or SSD Amounts
        # We start with section C because, in situations where someone receives both a pension and taxable
        # Social Security, Section A requires information from Section C to be completed.
        eligible_ss_or_ssd = person(
            "mo_pension_and_ss_or_ssd_deduction_section_c", period
        )
        tax_unit_eligible_ss_or_ssd = tax_unit.sum(eligible_ss_or_ssd)
        # Section A, Public Pension Amounts
        # TODO:
        # unclear reference to "See instructions if Line 3 of Section C is more than $0" here: https://dor.mo.gov/forms/MO-A_2021.pdf#page=3
        public_pensions = person(
            "mo_pension_and_ss_or_ssd_deduction_section_a", period
        )
        tax_unit_eligible_total_public_pensions = tax_unit.sum(public_pensions)
        # Section B, Private Pension Amounts
        total_private_pensions = person(
            "mo_pension_and_ss_or_ssd_deduction_section_b", period
        )
        tax_unit_total_private_pensions = tax_unit.sum(total_private_pensions)
        return (
            tax_unit_total_private_pensions
            + tax_unit_eligible_total_public_pensions
            + tax_unit_eligible_ss_or_ssd
        )



from policyengine_us.model_api import *


class va_federal_state_employees_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Virginia federal state employees subtraction"
    defined_for = StateCode.VA
    unit = USD
    definition_period = YEAR
    reference = (
        "https://law.lis.virginia.gov/vacodefull/title58.1/chapter3/article2/"
    )

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        p = parameters(
            period
        ).gov.states.va.tax.income.subtractions.disability_income

        employment_income = person("irs_employment_income", period)
        state_or_federal_salary = person("state_or_federal_salary", period)
        subtractable_federal_state_salary = where(
            employment_income > p.amount, 0, state_or_federal_salary
        )
        head = person("is_tax_unit_head", period)
        spouse = person("is_tax_unit_spouse", period)
        is_head_or_spouse = head | spouse
        return tax_unit.sum(
            subtractable_federal_state_salary * is_head_or_spouse
        )



from policyengine_us.model_api import *


class ut_state_tax_refund(Variable):
    value_type = float
    entity = TaxUnit
    label = "Utah state tax refund"
    unit = USD
    documentation = "Form TC-40, line 7"
    definition_period = YEAR
    defined_for = StateCode.UT

    adds = ["salt_refund_last_year"]



from policyengine_us.model_api import *


class has_all_usda_elderly_disabled(Variable):
    value_type = bool
    entity = SPMUnit
    definition_period = YEAR
    documentation = "Whether the SPM unit's members all meet USDA definitions of elderly or disabled"
    label = "Has all USDA elderly or disabled people"
    # NB: This isn't used in SNAP directly, but it is used for BBCE (TANF).

    def formula(spm_unit, period, parameters):
        person = spm_unit.members
        elderly = person("is_usda_elderly", period)
        disabled = person("is_usda_disabled", period)
        return spm_unit.all(elderly | disabled)



from policyengine_us.model_api import *


class ma_tafdc_financial_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for Massachusetts Temporary Assistance for Families with Dependent Children (TAFDC) due to income"
    definition_period = MONTH
    reference = (
        "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-704-000"
    )
    defined_for = StateCode.MA

    def formula(spm_unit, period, parameters):
        total_countable_income = spm_unit(
            "ma_tafdc_applicable_income_for_financial_eligibility", period
        )
        payment_standard = spm_unit("ma_tafdc_payment_standard", period)

        return total_countable_income < payment_standard



from policyengine_us.model_api import *


class va_child_dependent_care_deduction_cdcc_limit(Variable):
    value_type = float
    entity = TaxUnit
    label = (
        "Federal CDCC-relevant care expense limit for Virginia tax purposes"
    )
    unit = USD
    definition_period = YEAR
    reference = "https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2021-760-instructions.pdf#page=29"
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        if period.start.year == 2021:
            instant_str = f"2020-01-01"
        else:
            instant_str = period
        p = parameters(instant_str).gov.irs.credits.cdcc
        capped_count_cdcc_eligible = tax_unit(
            "capped_count_cdcc_eligible", period
        )
        return p.max * capped_count_cdcc_eligible



from policyengine_us.model_api import *


class ms_itemized_deductions_joint(Variable):
    value_type = float
    entity = Person
    label = "Mississippi itemized deductions for joint couples"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.dor.ms.gov/sites/default/files/Forms/Individual/80100221.pdf#page=15"
        "https://www.dor.ms.gov/sites/default/files/Forms/Individual/80108228.pdf, # Line 7"
        "https://casetext.com/statute/mississippi-code-1972/title-27-taxation-and-finance/chapter-7-income-tax-and-withholding/article-1-income-tax/section-27-7-17-deductions-allowed?__cf_chl_rt_tk=8Kelu8kHpIXTp_FnAJLHvqa7rtrZYE1U.NAeBM8L.Nc-1692990420-0-gaNycGzNEmU"
    )
    defined_for = StateCode.MS

    def formula(person, period, parameters):
        deductions = person.tax_unit("ms_itemized_deductions_unit", period)
        is_head = person("is_tax_unit_head", period)
        return deductions * is_head



from policyengine_us.model_api import *


class ne_income_tax_before_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "NE income tax before refundable credits"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/past-year/2021/511-Pkt-2021.pdf"
        "https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-Pkt.pdf"
    )
    defined_for = StateCode.NE

    def formula(tax_unit, period, parameters):
        itax_before_credits = tax_unit("ne_income_tax_before_credits", period)
        nonrefundable_credits = tax_unit("ne_nonrefundable_credits", period)
        ne_amount = max_(0, itax_before_credits - nonrefundable_credits)
        # remaining calculations follow the Federal Tax Liability Worksheet
        ne_additions = tax_unit("ne_additions", period)
        ne_agi_subtractions = tax_unit("ne_agi_subtractions", period)
        add_less_sub = ne_additions - ne_agi_subtractions
        p = parameters(period).gov.states.ne.tax.income
        adjustment = add_less_sub < p.credits.nonrefundable_adjust_limit
        us_amt = tax_unit("income_tax_before_credits", period)
        us_amount = where(adjustment, us_amt, ne_amount)
        return min_(ne_amount, us_amount)



from policyengine_us.model_api import *


class nm_medical_care_expense_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Mexico medical care expense deduction"
    unit = USD
    definition_period = YEAR
    reference = (
        # 7-2-35. DEDUCTION – UNREIMBURSED OR UNCOMPENSATED MEDICAL CARE EXPENSES
        "https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/856ebf4b-3814-49dd-8631-ebe579d6a42b/Personal%20Income%20Tax.pdf#page=249",
        # Tax Form Instructions Page PIT-1-26 LINE 16. Medical Care Expense Deduction
        "https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf#page=31",
        # 7-2-37. Deduction; unreimbursed or uncompensated medical care expenses.
        "https://nmonesource.com/nmos/nmsa/en/item/4340/index.do#!fragment/zoupio-_Toc140503888/BQCwhgziBcwMYgK4DsDWszIQewE4BUBTADwBdoAvbRABwEtsBaAfX2zgEYAWABgFYeAZgAcogJQAaZNlKEIARUSFcAT2gBydRIiEwuBIuVrN23fpABlPKQBCagEoBRADKOAagEEAcgGFHE0jAAI2hSdjExIA",
    )
    defined_for = StateCode.NM

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.nm.tax.income.deductions.medical_care_expense
        # the deduction amount is based on filing status, agi, and eligible expenses.
        filing_status = tax_unit("filing_status", period)
        statuses = filing_status.possible_values
        agi = tax_unit("adjusted_gross_income", period)
        expenses = add(
            tax_unit,
            period,
            ["medical_out_of_pocket_expenses"],
        )
        # Use `right=True` to reflect "over ... but not over ...".
        rate = select(
            [
                filing_status == statuses.SINGLE,
                filing_status == statuses.JOINT,
                filing_status == statuses.HEAD_OF_HOUSEHOLD,
                filing_status == statuses.SEPARATE,
                filing_status == statuses.SURVIVING_SPOUSE,
            ],
            [
                p.single.calc(agi, right=True),
                p.joint.calc(agi, right=True),
                p.head_of_household.calc(agi, right=True),
                p.separate.calc(agi, right=True),
                p.surviving_spouse.calc(agi, right=True),
            ],
        )
        return expenses * rate



from policyengine_us.model_api import *


class mi_subtractions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Michigan taxable income subtractions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/Schedule-1.pdf",
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf",
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/MI-1040.pdf",
    )
    defined_for = StateCode.MI

    adds = "gov.states.mi.tax.income.subtractions"



from policyengine_us.model_api import *


class in_healthcare_sharing_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Indiana healthcare sharing ministry deduction"
    unit = USD
    definition_period = YEAR
    documentation = "Healthcare sharing expenses paid by a qualified individual for membership in a health care sharing ministry allowable for deduction in Indiana."
    reference = "https://iga.in.gov/laws/2024/ic/titles/6#6-3-2-28"
    defined_for = StateCode.IN



from policyengine_us.model_api import *


class ca_state_supplement_aged_disabled_count(Variable):
    value_type = int
    entity = SPMUnit
    label = "California SSI state supplement aged or disabled count"
    definition_period = MONTH
    defined_for = StateCode.CA
    reference = "https://leginfo.legislature.ca.gov/faces/codes_displaySection.xhtml?lawCode=WIC&sectionNum=12200"

    def formula(spm_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.ca.cdss.state_supplement.payment_standard
        person = spm_unit.members
        # Aged or disabled amount
        is_disabled = person("is_disabled", period)
        age = person("monthly_age", period)
        is_aged = age >= p.aged_or_disabled.age_threshold
        aged_or_disabled = is_aged | is_disabled
        eligible = person("ca_state_supplement_eligible_person", period)
        return spm_unit.sum(aged_or_disabled * eligible)



from policyengine_us.model_api import *


class nc_scca(Variable):
    value_type = float
    entity = SPMUnit
    label = "North Carolina Subsidized Child Care Assistance Program"
    unit = USD
    definition_period = MONTH
    defined_for = "nc_scca_entry_eligible"

    def formula(spm_unit, period, parameters):
        person = spm_unit.members
        age_eligible = person("nc_scca_child_age_eligible", period)

        # Collect the market rates only for eligible children
        total_market_rate = spm_unit.sum(
            person("nc_scca_market_rate", period) * age_eligible
        )

        parent_fee = spm_unit("nc_scca_parent_fee", period)
        uncapped_amount = max_(total_market_rate - parent_fee, 0)
        childcare_expenses = spm_unit(
            "spm_unit_pre_subsidy_childcare_expenses", period
        )
        return min_(childcare_expenses, uncapped_amount)
